from typing import Optional
from typing import Union
from typing import Sequence
from typing import Any

from packaging import version
from collections import defaultdict as ddict

import rdkit
from rdkit.Chem import rdDepictor
from rdkit.Chem import rdMolAlign

import pandas as pd
import datamol as dm


from .types import Mol


def compute_2d_coords(mol: Mol, copy: bool = True, verbose: bool = False) -> Mol:
    """Compute 2D coordinates for a molecule.

    Args:
        mol: A molecule.
        copy: Whether to copy the molecule.
    """
    if copy:
        mol = dm.copy_mol(mol)

    with dm.without_rdkit_log(enable=not verbose):
        rdDepictor.Compute2DCoords(mol)

    return mol


def template_align(
    mol: Union[str, Mol],
    template: Optional[Union[str, Mol]] = None,
    copy: bool = True,
    use_depiction: bool = True,
    remove_confs: bool = True,
    auto_select_coord_gen: bool = False,
) -> Optional[Mol]:
    """Align an input molecule to a template. If the template is not provided then the input molecule is
    returned.

    Args:
        mol: A molecule.
        template: Template to align to.
        copy: whether to copy the molecule before aligning it.
        use_depiction: Whether to use the depiction API or use MolAlign
            The main difference is around how 3D information is handled, but also, because the depiction API
            will emphasize the atoms that do not match, whereas AlignMol will not.
        remove_confs: Whether to remove all conformation in the input molecule first.
            You can set this to true when not using depiction
        auto_select_coord_gen: Whether to automatically select the coordinate generation method.

    Returns:
        mol: aligned molecule (dm.Mol). None if initial mol argument is undefined or invalid.
    """

    if isinstance(mol, str):
        _mol = dm.to_mol(mol)
    elif copy:
        _mol = dm.copy_mol(mol)
    else:
        _mol = mol

    if _mol is None:
        return None

    if isinstance(template, str):
        _template = dm.to_mol(template)
    elif copy:
        _template = dm.copy_mol(template)
    else:
        _template = template

    if _template is None:
        return _mol

    if remove_confs:
        _mol.RemoveAllConformers()

    # EN: to make this more general and robust, we need to first check whether the template
    # has 2D coordinates or not. If it does not, we need to compute them.
    # This is a very rare edge case if requests are comming from the UI, but better to check than not xD
    if _template.GetNumConformers() == 0:
        _template = compute_2d_coords(_template)

    # EN: now we can align the molecule to the template
    # but first, we should avoid MCS as much as possible, because it's expensive
    # so if the template is a subgraph of the molecule, no need to perform any MCS
    # Another reason for this, is to avoid inconsistency in alignment between molecules that are
    # supergraph of the template vs molecules that are subgraph of the template.
    pattern = _template
    if not _mol.HasSubstructMatch(_template):
        pattern = None
        mcs_smarts = dm.find_mcs([_mol, _template])

        if mcs_smarts is not None:
            pattern = dm.from_smarts(mcs_smarts)

    if pattern is not None:
        if auto_select_coord_gen:
            rdDepictor.SetPreferCoordGen(use_depiction)

        # we would need to compute 2d coordinates for the molecules if it doesn't have any
        if _mol.GetNumConformers() == 0:
            _mol = compute_2d_coords(_mol)

        if use_depiction:
            rdDepictor.GenerateDepictionMatching2DStructure(
                _mol,
                reference=_template,
                refPatt=pattern,
                acceptFailure=True,
                allowRGroups=True,
            )
        else:
            query_match = _mol.GetSubstructMatch(pattern)
            template_match = _template.GetSubstructMatch(pattern)
            rdMolAlign.AlignMol(_mol, _template, atomMap=list(zip(query_match, template_match)))

    return _mol


def auto_align_many(
    mols: Union[Sequence[Mol], pd.Series],
    partition_method: str = "anon-scaffold",
    copy: bool = True,
    cluster_cutoff: float = 0.7,
    allow_r_groups: bool = True,
    **kwargs: Any,
):
    """Partition a list of molecules into clusters sharing common scaffold of common core,
    then align the molecules to that common core. This function will compute the list of
    smiles/smarts representative of each cluster first.

    The returned molecules will have two properties associated to them:

    - `dm.auto_align_many.cluster_id`: the cluster id of the molecule.
    - `dm.auto_align_many.core`: the smiles/smarts of the core of the cluster.

    Args:
        mols: A list of molecules to auto align.
        partition_method: Partition method to use:

            - 'scaffold': Cluster molecules by Murcko scaffold.
            - 'strip-scaffold': Cluster molecules by Murcko scaffold, but remove all atoms not
                in the core.
            - 'anon-scaffold': Cluster molecules by Murcko scaffold, but making it
                generic including the bonds.
            - 'anongraph-scaffold': Cluster molecules by Murcko scaffold, but making it
                generic but keeping the bond order informations.
            - 'cluster': Cluster the molecules using Butina frm RDKit with `dm.cluster_mols`.
            Cautious as the method 'cluster' is very sensitive to the cutoff.

        copy: Whether to copy the molecules before aligning them.
        cluster_cutoff: Optional cluster cutoff.
        allow_r_groups: Optional, if True, terminal dummy atoms in the
                        reference are ignored if they match an implicit hydrogen in the
                        molecule, and a constrained depiction is still attempted
        **kwargs: Additional arguments to pass to clustering method
    """

    if copy:
        mols = [dm.copy_mol(mol) for mol in mols]

    mol_groups = ddict(list)  # map scaffold index to list of unique molecules
    scaffold_mols = {}

    if partition_method.endswith("scaffold"):
        scaffolds = [dm.to_scaffold_murcko(m) for m in mols]
        scaffolds_ids = [dm.to_smiles(x) for x in scaffolds]

        if partition_method.startswith("strip-"):
            scaffolds = [dm.strip_mol_to_core(x) for x in scaffolds]
            scaffolds_ids = [dm.to_smiles(x) for x in scaffolds]

        elif partition_method.startswith("anongraph-"):
            scaffolds = [dm.make_scaffold_generic(s, include_bonds=True) for s in scaffolds]
            scaffolds_ids = [dm.to_smiles(x) for x in scaffolds]

        elif partition_method.startswith("anon-"):
            scaffolds = [dm.make_scaffold_generic(s, include_bonds=False) for s in scaffolds]
            scaffolds_ids = [dm.to_smiles(x) for x in scaffolds]

        for i, s in enumerate(scaffolds_ids):
            mol_groups[s].append(i)
            scaffolds[i] = compute_2d_coords(scaffolds[i])
            scaffold_mols[s] = scaffolds[i]

    elif partition_method == "cluster":
        # partition is cluster, first compute molecule clusters
        clusters, mol_clusters = dm.cluster_mols(mols, cutoff=cluster_cutoff, **kwargs)

        # now compute the mcs for each clusters
        cluster_mcs = [
            (dm.find_mcs(mol_cluster) if len(mol_cluster) > 1 else dm.to_smiles(mol_cluster[0]))
            for mol_cluster in mol_clusters
        ]
        scaffolds_ids = [cluster_mcs[cluster_id] for cluster_id, _ in enumerate(clusters)]

        for i, s in enumerate(scaffolds_ids):
            mol_groups[s].extend(clusters[i])

        for x in scaffolds_ids:
            core = None
            if x is not None:
                core = dm.from_smarts(x)
                core = compute_2d_coords(core)
            scaffold_mols[x] = core

    else:
        raise ValueError(f"Unknown partition method: {partition_method}")

    # now we match each molecule to the scaffold and align them
    # note that the molecule object will be modified in place in the list
    for cluster_id, (core, mols_ids) in enumerate(mol_groups.items()):
        core_mol = scaffold_mols[core]

        for mol_id in mols_ids:
            mol = mols[mol_id]

            if core_mol is not None:
                # Only pass allowRGroups if current
                # rdkit version is >= 2021_03_1
                # ref https://github.com/rdkit/rdkit/pull/3811
                allowRGroups = (
                    {"allowRGroups": allow_r_groups}
                    if version.parse(rdkit.__version__) >= version.parse("2021.03.1")
                    else {}
                )
                rdDepictor.GenerateDepictionMatching2DStructure(
                    mol,
                    reference=core_mol,
                    acceptFailure=True,
                    **allowRGroups,
                )

            # Add some props to the mol so the user can retrieve the groups from
            # it later.
            props = {}
            props["dm.auto_align_many.cluster_id"] = cluster_id
            props["dm.auto_align_many.core"] = core
            dm.set_mol_props(mol, props)

    # EN: you can discard the mol_groups (or keep it and match the values
    # to molecular line notation, so you will not have to reocompute the above)
    # and convert the mols into cxsmiles if you want
    # return mols, mol_groups

    return mols



from typing import Union
from typing import Optional
from typing import List
from typing import Sequence
from typing import IO
from typing import Any
from typing import cast
from typing import overload
from typing import Literal

import os
import io
import tempfile
import pathlib

from rdkit.Chem import PandasTools
from rdkit.Chem import rdmolfiles

import pandas as pd
import fsspec
import fsspec.utils

import datamol as dm

from .types import Mol


def read_csv(
    urlpath: Union[str, os.PathLike, IO],
    smiles_column: Optional[str] = None,
    mol_column: str = "mol",
    **kwargs: Any,
) -> pd.DataFrame:
    """Read a CSV file.

    Args:
        urlpath: Path to a file or a file-like object. Path can be remote or local.
        smiles_column: Use this column to build a mol column.
        mol_column: Name to give to the mol column. If not None a mol column will be build.
            Avoid when loading a very large file.
        **kwargs: Arguments to pass to `pd.read_csv()`.

    Returns:
        df: a `pandas.DataFrame`
    """

    df = pd.read_csv(urlpath, **kwargs)
    df = cast(pd.DataFrame, df)

    if smiles_column is not None:
        PandasTools.AddMoleculeColumnToFrame(df, smiles_column, mol_column)

    return df


def read_excel(
    urlpath: Union[str, os.PathLike, IO],
    sheet_name: Optional[Union[str, int, list]] = 0,
    smiles_column: Optional[str] = None,
    mol_column: str = "mol",
    **kwargs: Any,
) -> pd.DataFrame:
    """Read an excel file.

    Args:
        urlpath: Path to a file or a file-like object. Path can be remote or local.
        sheet_name: see `pandas.read_excel()` doc.
        mol_column: Name to give to the mol column. If not None a mol column will be build.
            Avoid when loading a very large file.
        mol_column: name to give to the mol column.
        **kwargs: Arguments to pass to `pd.read_excel()`.

    Returns:
        df: a `pandas.DataFrame`
    """

    df = pd.read_excel(urlpath, sheet_name=sheet_name, **kwargs)
    df = cast(pd.DataFrame, df)

    if smiles_column is not None:
        PandasTools.AddMoleculeColumnToFrame(df, smiles_column, mol_column)

    return df


def _get_supplier_mols(
    supplier: rdmolfiles.ForwardSDMolSupplier,
    max_num_mols: Optional[int],
) -> List[dm.Mol]:
    """Given a supplier, read the molecules until we reach the `max_num_mols` limit.
    Useful when reading SDF files.
    """
    if max_num_mols is None:
        mols = list(supplier)
    else:
        mols = []
        for _ in range(max_num_mols):
            try:
                mols.append(next(supplier))
            except StopIteration:
                break
    return mols


@overload
def read_sdf(
    urlpath: Union[str, os.PathLike, IO],
    sanitize: bool = ...,
    as_df: Literal[False] = ...,
    smiles_column: Optional[str] = ...,
    mol_column: Optional[str] = ...,
    include_private: bool = ...,
    include_computed: bool = ...,
    strict_parsing: bool = ...,
    remove_hs: bool = ...,
    max_num_mols: Optional[int] = ...,
    discard_invalid: bool = ...,
    n_jobs: Optional[int] = ...,
) -> List[Mol]:
    ...


@overload
def read_sdf(
    urlpath: Union[str, os.PathLike, IO],
    sanitize: bool = ...,
    as_df: Literal[True] = ...,
    smiles_column: Optional[str] = ...,
    mol_column: Optional[str] = ...,
    include_private: bool = ...,
    include_computed: bool = ...,
    strict_parsing: bool = ...,
    remove_hs: bool = ...,
    max_num_mols: Optional[int] = ...,
    discard_invalid: bool = ...,
    n_jobs: Optional[int] = ...,
) -> pd.DataFrame:
    ...


@overload
def read_sdf(
    urlpath: Union[str, os.PathLike, IO],
    sanitize: bool = ...,
    as_df: bool = ...,
    smiles_column: Optional[str] = ...,
    mol_column: Optional[str] = ...,
    include_private: bool = ...,
    include_computed: bool = ...,
    strict_parsing: bool = ...,
    remove_hs: bool = ...,
    max_num_mols: Optional[int] = ...,
    discard_invalid: bool = ...,
    n_jobs: Optional[int] = ...,
) -> Union[List[Mol], pd.DataFrame]:
    ...


def read_sdf(
    urlpath: Union[str, os.PathLike, IO],
    sanitize: bool = True,
    as_df: bool = False,
    smiles_column: Optional[str] = "smiles",
    mol_column: Optional[str] = None,
    include_private: bool = False,
    include_computed: bool = False,
    strict_parsing: bool = True,
    remove_hs: bool = True,
    max_num_mols: Optional[int] = None,
    discard_invalid: bool = True,
    n_jobs: Optional[int] = 1,
) -> Union[List[Mol], pd.DataFrame]:
    """Read an SDF file.

    Note: This function is meant to be used with dataset that fit _in-memory_.
    For a more advanced usage we suggest you to use directly `Chem.ForwardSDMolSupplier`.

    Args:
        urlpath: Path to a file or a file-like object. Path can be remote or local.
        sanitize: Whether to sanitize the molecules.
        as_df: Whether to return a list mol or a pandas DataFrame.
        smiles_column: Name of the SMILES column. Only relevant if `as_df` is True.
        mol_column: Name of the mol column. Only relevant if `as_df` is True.
        include_private: Include private properties in the columns.  Only relevant if
            `as_df` is True.
        include_computed: Include computed properties in the columns.  Only relevant if
            `as_df` is True.
        strict_parsing: If set to false, the parser is more lax about correctness of the contents.
        remove_hs: Whether to remove the existing hydrogens in the SDF files.
        max_num_mols: Maximum number of molecules to read from the SDF file. Read all by default when set
            to `None`.
        discard_invalid: Discard the molecules that failed to be read correctly. Otherwise,
            invalid molecules will be loaded as `None`.
        n_jobs: Optional number of jobs for parallelization of `to_df`. Leave to 1 for no
            parallelization. Set to -1 to use all available cores. Only relevant is `as_df` is True
    """

    # File-like object
    if isinstance(urlpath, io.IOBase):
        supplier = rdmolfiles.ForwardSDMolSupplier(
            urlpath,
            sanitize=sanitize,
            strictParsing=strict_parsing,
            removeHs=remove_hs,
        )
        mols = _get_supplier_mols(supplier, max_num_mols)

    # Regular local or remote paths
    else:
        with fsspec.open(urlpath, compression="infer") as f:
            supplier = rdmolfiles.ForwardSDMolSupplier(
                f,
                sanitize=sanitize,
                strictParsing=strict_parsing,
                removeHs=remove_hs,
            )
            mols = _get_supplier_mols(supplier, max_num_mols)

    # Discard None values
    if discard_invalid:
        mols = [mol for mol in mols if mol is not None]

    # Convert to dataframe
    if as_df:
        return dm.to_df(
            mols,
            smiles_column=smiles_column,
            mol_column=mol_column,
            include_private=include_private,
            include_computed=include_computed,
            n_jobs=n_jobs,
        )  # type: ignore

    return mols


def to_sdf(
    mols: Union[Mol, Sequence[Mol], pd.DataFrame],
    urlpath: Union[str, os.PathLike, IO],
    smiles_column: Optional[str] = "smiles",
    mol_column: Optional[str] = None,
):
    """Write molecules to a file.

    Args:
        mols: a dataframe, a molecule or a list of molecule.
        urlpath: Path to a file or a file-like object. Path can be remote or local.
        smiles_column: Column name to extract the molecule.
        mol_column: Column name to extract the molecule. It takes
            precedence over `smiles_column`.
    """

    if isinstance(mols, pd.DataFrame):
        mols = dm.from_df(mols, smiles_column=smiles_column, mol_column=mol_column)

    elif isinstance(mols, Mol):
        mols = [mols]

    # Filter out None values
    mols = [mol for mol in mols if mol is not None]

    # File-like object
    if isinstance(urlpath, io.IOBase):
        writer = rdmolfiles.SDWriter(urlpath)
        for mol in mols:
            writer.write(mol)
        writer.close()

    # Regular local or remote paths
    else:
        with fsspec.open(urlpath, mode="w") as f:
            writer = rdmolfiles.SDWriter(f)
            for mol in mols:
                writer.write(mol)
            writer.close()


def read_mol2file(
    urlpath: Union[str, os.PathLike, IO],
    sanitize: bool = True,
    cleanup_substructures: bool = True,
    remove_hs: bool = True,
    fail_if_invalid: bool = False,
) -> List[Mol]:
    """Read a Mol2 File

    Args:
        urlpath: Path to a file or a file-like object. Path can be remote or local.
        sanitize: Whether to sanitize the molecules.
        remove_hs: Whether to remove the existing hydrogens in the SDF files.
        cleanup_substructures: Whether to clean up substructure in the Mol2 Files.
        fail_if_invalid: If set to true, the parser will raise an exception if the molecule is invalid
            instead of returning None.
    """

    block = []
    mols = []
    with fsspec.open(urlpath, compression="infer") as f:
        fReadLines = f.readlines()
        # reversing due to ambiguous end line for mol2 files
        fReadLines.reverse()
        for line in fReadLines:
            # ignores any header info
            if b"#" not in line:
                block.append(str(line, "utf-8"))
            # since reversed, this is the 'end' a mol2
            if b"@<TRIPOS>MOLECULE" in line:
                block.reverse()
                mol2block = ",".join(block).replace(",", "")
                mol = rdmolfiles.MolFromMol2Block(
                    mol2block,
                    sanitize=sanitize,
                    removeHs=remove_hs,
                    cleanupSubstructures=cleanup_substructures,
                )
                if mol is None and fail_if_invalid:
                    raise ValueError(f"Invalid molecule: {mol2block}")
                mols.append(mol)
                block = []

    mols.reverse()
    return mols


def read_molblock(
    molblock: str,
    sanitize: bool = True,
    strict_parsing: bool = True,
    remove_hs: bool = True,
    fail_if_invalid: bool = False,
) -> Optional[dm.Mol]:
    """Read a Mol block.

    Note that potential molecule properties are **not** read.

    Args:
        molblock: String containing the Mol block.
        sanitize: Whether to sanitize the molecules.
        strict_parsing: If set to false, the parser is more lax about correctness of the contents.
        remove_hs: Whether to remove the existing hydrogens in the SDF files.
        fail_if_invalid: If set to true, the parser will raise an exception if the molecule is invalid
            instead of returning None.
    """

    mol = rdmolfiles.MolFromMolBlock(
        molblock,
        sanitize=sanitize,
        removeHs=remove_hs,
        strictParsing=strict_parsing,
    )

    if mol is None and fail_if_invalid:
        raise ValueError(f"Invalid molecule: {molblock}")

    return mol


def to_molblock(
    mol: Mol,
    include_stereo: bool = True,
    conf_id: int = -1,
    kekulize: bool = True,
    force_V3000: bool = False,
):
    """Convert a molecule to a mol block string.

    Note that any molecule properties are lost.

    Args:
        mol: A molecule.
        include_stereo: Toggles inclusion of stereochemical information in the output.
        conf_id: Selects which conformation to output.
        kekulize: Triggers kekulization of the molecule before it's written,
            as suggested by the MDL spec.
        force_V3000: Force generation a V3000 mol block (happens automatically
            with more than 999 atoms or bonds).
    """

    molblock = rdmolfiles.MolToMolBlock(
        mol,
        includeStereo=include_stereo,
        confId=conf_id,
        kekulize=kekulize,
        forceV3000=force_V3000,
    )

    return molblock


def read_pdbblock(
    molblock: str,
    sanitize: bool = True,
    remove_hs: bool = True,
    flavor: int = 0,
    proximity_bonding: bool = True,
) -> dm.Mol:
    """Read a PDB string block.

    Args:
        molblock: String containing the Mol block.
        sanitize: Whether to sanitize the molecules.
        remove_hs: Whether to remove the existing hydrogens in the SDF files.
        flavor: RDKit flavor options.
        proximity_bonding: Whether to toggles automatic proximity bonding.
    """

    mol = rdmolfiles.MolFromPDBBlock(
        molblock,
        sanitize=sanitize,
        removeHs=remove_hs,
        flavor=flavor,
        proximityBonding=proximity_bonding,
    )
    return mol


def to_pdbblock(mol: Mol, conf_id: int = -1) -> str:
    """Convert a molecule to a PDB string block.

    Args:
        mol: A molecule.
        conf_id: Selects which conformation to use.
    """
    molblock = rdmolfiles.MolToPDBBlock(mol, confId=conf_id)
    return molblock


def read_pdbfile(
    urlpath: Union[str, os.PathLike],
    sanitize: bool = True,
    remove_hs: bool = True,
    flavor: int = 0,
    proximity_bonding: bool = True,
) -> Mol:
    """Read a PDB file.

    Args:
        urlpath: Path to a file or a file-like object. Path can be remote or local.
        sanitize: Whether to sanitize the molecules.
        remove_hs: Whether to remove the existing hydrogens in the SDF files.
        flavor: RDKit flavor options.
        proximity_bonding: Whether to toggles automatic proximity bonding.

    Returns:
        mol: a molecule
    """

    with fsspec.open(urlpath, "r") as f:
        f = cast(IO, f)
        mol = read_pdbblock(
            f.read(),
            sanitize=sanitize,
            remove_hs=remove_hs,
            flavor=flavor,
            proximity_bonding=proximity_bonding,
        )
    return mol


def to_pdbfile(
    mol: Mol,
    urlpath: Union[str, os.PathLike],
    conf_id: int = -1,
):
    """Save a molecule to a PDB file.

    Args:
        mol: A molecule.
        urlpath: Path to a file or a file-like object. Path can be remote or local.
        conf_id: Selects which conformation to use.
    """
    molblock = to_pdbblock(mol, conf_id=conf_id)
    with fsspec.open(urlpath, "w") as f:
        f = cast(IO, f)
        f.write(molblock)


def to_smi(
    mols: Sequence[Mol],
    urlpath: Union[str, os.PathLike, IO],
    error_if_empty: bool = False,
):
    """Save a list of molecules in an `.smi` file.

    Note: We **strongly** recommend you to use `dm.to_csv` instead
    of `dm.to_smi` since `.smi` files are CSV-like format. The only difference are the
    default settings which changes:

    - The default separator is a space ` ` instead of a comma `,`.
    - The headers of the column are not included.

    By modifying the args of `dm.to_csv()`, you will be able to save a SMI compatible file.

    Args:
        mols: a list of molecules.
        urlpath: Path to a file or a file-like object. Path can be remote or local.
        error_if_empty: whether to raise and error if the input list is empty.
    """

    if len(mols) == 0 and error_if_empty:
        raise ValueError("The list of mols/smiles provided is empty.")

    # Filter out None values
    mols = [mol for mol in mols if mol is not None]

    # File-like object
    if isinstance(urlpath, io.IOBase):
        writer = rdmolfiles.SmilesWriter(urlpath, includeHeader=False, nameHeader="")
        for mol in mols:
            writer.write(mol)
        writer.close()

    # Regular local or remote paths
    else:
        with fsspec.open(urlpath, "w") as f:
            writer = rdmolfiles.SmilesWriter(f, includeHeader=False, nameHeader="")
            for mol in mols:
                writer.write(mol)
            writer.close()


def read_smi(
    urlpath: Union[str, pathlib.Path, io.IOBase, fsspec.core.OpenFile],
) -> Sequence[Mol]:
    """Read a list of smiles from am `.smi` file.

    Note: We **strongly** recommend you to use `dm.read_csv` or `pandas.read_csv` instead
    of `dm.read_smi` since `.smi` files are CSV-like format. The only difference are the
    default settings which changes:

    - The default separator is a space ` ` instead of a comma `,`.
    - The headers of the column are not included.

    By modifying the args of `dm.read_csv()`, you will be able to read an `.smi` files.

    Args:
        urlpath: Path to a file or a file-like object. Path can be remote or local.
    """

    active_path = urlpath

    # NOTE(hadim): the temporary local file copy
    # is because `SmilesMolSupplier` does not support
    # using file-like object, only path.

    # Copy to a local temporary path if the path is a remote one.
    if not fsspec.utils.can_be_local(str(urlpath)):
        active_path = pathlib.Path(tempfile.mkstemp()[1])
        dm.utils.fs.copy_file(urlpath, active_path, force=True)

    # Read the molecules
    supplier = rdmolfiles.SmilesMolSupplier(str(active_path), titleLine=0)
    mols = [mol for mol in supplier if mol is not None]

    # Delete the local temporary path
    if not fsspec.utils.can_be_local(str(urlpath)):
        pathlib.Path(str(active_path)).unlink()

    return mols


def to_xlsx(
    mols: Union[Mol, Sequence[Mol], pd.DataFrame],
    urlpath: Union[str, os.PathLike],
    smiles_column: Optional[str] = "smiles",
    mol_column: str = "mol",
    mol_size: List[int] = [300, 300],
):
    """Write molecules to an Excel file with a molecule column as an RDKit rendered
    image.

    Args:
        mols: a dataframe, a molecule or a list of molecule.
        urlpath: Path to a file or a file-like object. Path can be remote or local.
        smiles_column: Column name to extract the molecule.
        mol_column: Column name to extract the molecule. It takes
            precedence over `smiles_column`.
            Column name to write the RDKit rendered image. If none,
            the molecule images are not written.
    """

    if isinstance(mols, Mol):
        mols = [mols]

    if isinstance(mols, Sequence):
        mols = [mol for mol in mols if mol is not None]
        mols = dm.to_df(mols, smiles_column=smiles_column, mol_column=mol_column)

    if mols is None or mols.empty:  # type: ignore
        raise ValueError("No molecules to write")

    with fsspec.open(urlpath, mode="wb") as f:
        PandasTools.SaveXlsxFromFrame(mols, f, molCol=mol_column, size=mol_size)



import pytest

import multiprocessing
import platform

# NOTE(hadim): since the tests are now run in parallel with pytest-xdist
# the below test checking for the import duration is not relevant anymore.
# For now, we keep it while bumping the duration threshold to high values.

DATAMOL_MAX_IMPORT_DURATION = {}  # in seconds
DATAMOL_MAX_IMPORT_DURATION["default"] = 60
DATAMOL_MAX_IMPORT_DURATION["linux"] = 60
DATAMOL_MAX_IMPORT_DURATION["osx"] = 60
DATAMOL_MAX_IMPORT_DURATION["windows"] = 60


def _get_max_import_duration():
    if platform.system() == "Windows":
        return DATAMOL_MAX_IMPORT_DURATION["windows"]
    elif platform.system() == "Linux":
        return DATAMOL_MAX_IMPORT_DURATION["linux"]
    elif platform.system() == "Darwin":
        return DATAMOL_MAX_IMPORT_DURATION["osx"]
    else:
        return DATAMOL_MAX_IMPORT_DURATION["default"]


def _import_datamol_fn(queue):
    import datetime

    start = datetime.datetime.now()
    import datamol

    duration = datetime.datetime.now() - start
    duration = duration.total_seconds()
    queue.put(duration)


@pytest.mark.skip_platform("osx")
def test_datamol_import():
    context = multiprocessing.get_context(method="spawn")

    queue = context.Queue()
    p = context.Process(target=_import_datamol_fn, args=(queue,))
    p.start()
    duration = queue.get()
    p.join()
    p.close()

    print(duration)

    # Check datamol wasn't too fast to load
    # That should never happen and it's a simple sanity check
    # whether the `spawn` process worked correctly.
    assert duration > 1e-2, "datamol loaded too fast, something is wrong with the test."

    # Check datamol wasn't too long to load
    assert duration < _get_max_import_duration(), "datamol tooks too long to load."



from typing import cast
from typing import Union

import re
import operator

import datamol as dm
from rdkit import Chem

ATTACHMENT_POINT_TOKEN = "*"
ATTACHMENT_POINT_NUM_REGEXP = r"\[{}:?(\d*)\]".format(re.escape(ATTACHMENT_POINT_TOKEN))
ATTACHMENT_POINT_REGEXP = r"(?:{0}|\[{0}[^\]]*\])".format(re.escape(ATTACHMENT_POINT_TOKEN))
ATTACHMENT_POINT_NO_BRACKETS_REGEXP = r"(?<![:\[]){0}(?![:\]])".format(
    re.escape(ATTACHMENT_POINT_TOKEN)
)
ALL_POSSIBLE_ATTACHMENTS = r"\[(\d*){}:?(\d*)\]".format(re.escape(ATTACHMENT_POINT_TOKEN))


def add_brackets_to_attachment_points(smiles: str) -> str:
    """
    Adds brackets to the attachment points (if they don't have them).
    Example: "CC(C)CO*" to "CC(C)CO[*]"

    Args:
        smiles: A smiles string.

    Returns:
        A smiles string with brackets.
    """
    return re.sub(
        ATTACHMENT_POINT_NO_BRACKETS_REGEXP,
        "[{}]".format(ATTACHMENT_POINT_TOKEN),
        smiles,
    )


def convert_attach_to_isotope(
    mol_or_smiles: Union[dm.Mol, str],
    same_isotope: bool = False,
    as_smiles: bool = False,
) -> Union[dm.Mol, str]:
    """Convert attachment to isotope mapping.

    Examples: "O=C(NCc1cnc([*])c1)[*]" to  "O=C(NCc1cnc([1*])c1)[2*]"

    Args:
        mol_or_smiles: A Mol object or a smiles to be converted
        same_isotope: Whether convert to the same isotope.
            Example: "O=C(NCc1cnc([*])c1)[*]" to  "O=C(NCc1cnc([1*])c1)[1*]"

    Returns:
        Converted Mol object or SMILES.
    """
    mol = dm.to_mol(mol_or_smiles)
    smiles = dm.to_smiles(mol)
    smiles = cast(str, smiles)

    smiles = add_brackets_to_attachment_points(smiles)

    # reg matching seems to be the most effective
    subs_reg = r"[\g<1>{}]"
    if same_isotope:
        subs_reg = "[1{}]"

    smiles = re.sub(ATTACHMENT_POINT_NUM_REGEXP, subs_reg.format(ATTACHMENT_POINT_TOKEN), smiles)

    if as_smiles:
        return smiles
    return dm.to_mol(smiles)


def num_attachment_points(mol_or_smiles: Union[dm.Mol, str]) -> int:
    """
    Get the number of attachment point in the

    Args:
        mol_or_smiles: A Mol object or a smiles to be converted

    Returns:
        Number of attachment points of the given molecule.
    """
    if isinstance(mol_or_smiles, dm.Mol):
        mol = cast(dm.Mol, mol_or_smiles)
        n_points = len(
            [atom for atom in mol.GetAtoms() if atom.GetSymbol() == ATTACHMENT_POINT_TOKEN]
        )
    else:
        n_points = len(re.findall(ATTACHMENT_POINT_REGEXP, mol_or_smiles))

    return n_points


def open_attach_points(
    mol: dm.Mol,
    fix_atom_map: bool = False,
    bond_type: dm.BondType = dm.SINGLE_BOND,
) -> dm.Mol:
    """Compute attachment points on a molecule.
    This will highlight all valid attachment point on the current molecule instead.

    Args:
        mol: A Mol object to be processed.
        fix_atom_map: Whether fix the atom mapping of the molecule.
        bond_type: The bond type to be opened.

    Returns:
        Molecule with open attachment points
    """

    emol = Chem.rdchem.RWMol(dm.to_mol(mol))
    with dm.log.without_rdkit_log():
        atoms = [
            (a.GetIdx(), a)
            for a in emol.GetAtoms()
            if a.GetSymbol() != ATTACHMENT_POINT_TOKEN
            and a.GetImplicitValence() > 0
            and (not a.HasProp("_protected") or a.GetProp("_protected") != "1")
        ]
        atoms.sort(reverse=True, key=operator.itemgetter(0))

        for atom in atoms:
            new_atom = Chem.rdchem.Atom(ATTACHMENT_POINT_TOKEN)
            new_atom.SetAtomMapNum(1 if fix_atom_map else atom[0])
            new_index = emol.AddAtom(new_atom)
            emol.UpdatePropertyCache(strict=False)
            if bond_type is not None:
                emol.AddBond(atom[0], new_index, bond_type)
            else:
                emol.AddBond(atom[0], new_index)

    mol = dm.sanitize_mol(emol)
    return mol



from typing import Union
from typing import List
from typing import Optional
from typing import cast

import re

from loguru import logger

import pandas as pd

from rdkit import Chem
from rdkit.Chem import rdmolfiles
from rdkit.Chem import PandasTools

import selfies as sf

from .types import Mol

# NOTE(hadim): it's not possible to use `from .mol import ...` because of circular imports
import datamol as dm


def to_smiles(
    mol: Mol,
    canonical: bool = True,
    isomeric: bool = True,
    kekulize: bool = False,
    ordered: bool = False,
    explicit_bonds: bool = False,
    explicit_hs: bool = False,
    randomize: bool = False,
    cxsmiles: bool = False,
    allow_to_fail: bool = False,
    with_atom_indices: bool = False,
) -> Optional[str]:
    """Convert a mol to a SMILES.

    Args:
        mol: a molecule.
        canonical: if false no attempt will be made to canonicalize the molecule.
        isomeric: whether to include information about stereochemistry in the SMILES.
        kekulize: whether to return the kekule version of the SMILES.
        ordered: whether to force reordering of the atoms first.
        explicit_bonds: if true, all bond orders will be explicitly indicated in the output SMILES.
        explicit_hs: if true, all H counts will be explicitly indicated in the output SMILES.
        randomize: whether to randomize the generated smiles. Override `canonical`.
        cxsmiles: Whether to return a CXSMILES instead of a SMILES.
        allow_to_fail: Raise an error if the conversion to SMILES fails. Return None otherwise.
        with_atom_indices: Whether to add atom indices to the SMILES.
    """

    if ordered and canonical is False:
        mol = dm.reorder_atoms(mol)

    if randomize:
        mol = dm.randomize_atoms(mol)
        canonical = False

    if with_atom_indices:
        mol = dm.atom_indices_to_mol(mol, copy=True)

    smiles = None
    try:
        if cxsmiles:
            smiles = rdmolfiles.MolToCXSmiles(
                mol,
                isomericSmiles=isomeric,
                canonical=canonical,
                allBondsExplicit=explicit_bonds,
                allHsExplicit=explicit_hs,
                kekuleSmiles=kekulize,
            )

        else:
            smiles = rdmolfiles.MolToSmiles(
                mol,
                isomericSmiles=isomeric,
                canonical=canonical,
                allBondsExplicit=explicit_bonds,
                allHsExplicit=explicit_hs,
                kekuleSmiles=kekulize,
            )

    except Exception as e:
        if allow_to_fail:
            raise e

        return None

    return smiles


def to_selfies(mol: Union[str, Mol]) -> Optional[str]:
    """Convert a mol to SELFIES.

    Args:
        mol: a molecule or a SMILES.

    Returns:
        selfies: SELFIES string.
    """

    if isinstance(mol, Mol):
        mol = to_smiles(mol)

    if mol is None:
        return None
    selfies = sf.encoder(mol)

    if selfies == -1:
        return None

    return selfies


def from_selfies(selfies: str, as_mol: bool = False) -> Optional[Union[str, Mol]]:
    """Convert a SEFLIES to a smiles or a mol.

    Args:
        selfies: a selfies.
        as_mol (str, optional): whether to return a mol or a smiles.

    Returns:
        smiles or mol.
    """
    if selfies is None:
        return None

    smiles = sf.decoder(selfies)

    if as_mol and smiles is not None:
        return dm.to_mol(smiles)

    return smiles


def smiles_as_smarts(mol: Union[str, Mol], keep_hs: bool = True) -> Optional[str]:
    """Convert a smiles to a smarts if possible

    Args:
        mol: a molecule.
        keep_hs: Whether to keep hydrogen. This will increase the count of H atoms
            for atoms with attached hydrogens to create a valid smarts without further substitution allowed
            e.g. [H]-[CH]-[*] -> [H]-[CH2]-[*]

    Returns:
        smarts of the molecule
    """

    if isinstance(mol, str):
        mol = dm.to_mol(mol)

    if mol is None:
        return None

    # Change the isotope to 99
    for atom in mol.GetAtoms():  # type: ignore
        if keep_hs:
            s = sum(na.GetAtomicNum() == 1 for na in atom.GetNeighbors())
            if s:
                atom.SetNumExplicitHs(atom.GetTotalNumHs() + s)
        atom.SetIsotope(99)

    # Print out the smiles, all the atom attributes will be fully specified
    smarts = to_smiles(mol, isomeric=True, explicit_bonds=True)

    if smarts is None:
        return None

    # Remove the 99 isotope labels
    smarts = re.sub(r"\[99", "[", smarts)
    return smarts


def to_inchi(mol: Union[str, Mol]) -> Optional[str]:
    """Convert a mol to a standard Inchi.

    Args:
        mol: a molecule.
    """

    if isinstance(mol, str):
        mol = dm.to_mol(mol)

    if mol is None:
        return None

    inchi_val = Chem.MolToInchi(mol)
    if not inchi_val:
        return None
    return inchi_val


def to_inchi_non_standard(
    mol: Union[str, Mol],
    fixed_hydrogen_layer: bool = True,
    undefined_stereocenter: bool = True,
    reconnected_metal_layer: bool = True,
    tautomerism_keto_enol: bool = True,
    tautomerism_15: bool = True,
    options: Optional[List[str]] = None,
) -> Optional[str]:
    """Convert a mol to a non-standard Inchi.

    Note that turning all the flags to `False` will result in the standard Inchi.

    **Warning**: this function will return a **non-standard** Inchi. See
    https://www.inchi-trust.org/technical-faq-2 for details.

    It's important to not mix standard and non-standard InChi. If you don't know
    much about non-standard InChi, we highly recommend you to use the
    standard InChi with `dm.to_inchi()`.

    Args:
        mol: a molecule.
        fixed_hydrogen_layer: whether to include a fixed hydrogen layer (`/FixedH`).
        undefined_stereocenter: whether to include an undefined stereocenter layer (`/SUU`).
        reconnected_metal_layer: whether to include reconnected metals (`/RecMet`).
        tautomerism_keto_enol: whether to account tautomerism keto-enol (`/KET`).
        tautomerism_15: whether to account 1,5-tautomerism (`/15T`).
        options: More InchI options in a form of a list of string. Example:
            `["/SRel", "/AuxNone"]`.
    """

    if isinstance(mol, str):
        mol = dm.to_mol(mol)

    if mol is None:
        return None

    inchi_options = _process_inchi_options(
        fixed_hydrogen_layer=fixed_hydrogen_layer,
        undefined_stereocenter=undefined_stereocenter,
        reconnected_metal_layer=reconnected_metal_layer,
        tautomerism_keto_enol=tautomerism_keto_enol,
        tautomerism_15=tautomerism_15,
        options=options,
    )

    inchi_val = Chem.MolToInchi(mol, options=inchi_options)
    if not inchi_val:
        return None
    return inchi_val


def to_smarts(mol: Mol) -> Optional[str]:
    """Convert a mol to SMARTS format

    Args:
        mol: a molecule.
    """

    if mol is None:
        return None

    return Chem.MolToSmarts(mol)  # type: ignore


def to_inchikey(mol: Union[str, Mol]) -> Optional[str]:
    """Convert a mol to a standard InchiKey.

    Args:
        mol: a molecule
    """
    if isinstance(mol, str):
        mol = dm.to_mol(mol)

    if mol is None:
        return None

    inchikey = Chem.MolToInchiKey(mol)
    if not inchikey:
        return None
    return inchikey


def to_inchikey_non_standard(
    mol: Union[str, Mol],
    fixed_hydrogen_layer: bool = True,
    undefined_stereocenter: bool = True,
    reconnected_metal_layer: bool = True,
    tautomerism_keto_enol: bool = True,
    tautomerism_15: bool = True,
    options: Optional[List[str]] = None,
) -> Optional[str]:
    """Convert a mol to a non-standard InchiKey.

    Note that turning all the flags to `False` will result in the standard InchiKey.

    **Warning**: this function will return a **non-standard** InchiKey. See
    https://www.inchi-trust.org/technical-faq-2 for details.

    It's important to not mix standard and non-standard InChiKey. If you don't know
    much about non-standard InchiKey, we highly recommend you to use the
    standard InchiKey with `dm.to_inchikey()`.

    Args:
        mol: a molecule
        fixed_hydrogen_layer: whether to include a fixed hydrogen layer (`/FixedH`).
        undefined_stereocenter: whether to include an undefined stereocenter layer (`/SUU`).
        reconnected_metal_layer: whether to include reconnected metals (`/RecMet`).
        tautomerism_keto_enol: whether to account tautomerism keto-enol (`/KET`).
        tautomerism_15: whether to account 1,5-tautomerism (`/15T`).
        options: More InchI options in a form of a list of string. Example:
            `["/SRel", "/AuxNone"]`.
    """

    if isinstance(mol, str):
        mol = dm.to_mol(mol)

    if mol is None:
        return None

    inchi_options = _process_inchi_options(
        fixed_hydrogen_layer=fixed_hydrogen_layer,
        undefined_stereocenter=undefined_stereocenter,
        reconnected_metal_layer=reconnected_metal_layer,
        tautomerism_keto_enol=tautomerism_keto_enol,
        tautomerism_15=tautomerism_15,
        options=options,
    )

    inchikey = Chem.MolToInchiKey(mol, options=inchi_options)
    if not inchikey:
        return None
    return inchikey


def from_inchi(
    inchi: Optional[str],
    sanitize: bool = True,
    remove_hs: bool = True,
) -> Optional[Mol]:
    """Convert an InChi to a mol.

    Args:
        inchi: an inchi string.
        sanitize: do sanitize.
        remove_hs: do remove hs.

    Returns:
        mol
    """
    if inchi is None:
        return None

    return Chem.MolFromInchi(inchi, sanitize=sanitize, removeHs=remove_hs)


def from_smarts(smarts: Optional[str]) -> Optional[Mol]:
    """Convert a SMARTS string to a molecule

    Args:
        smarts: a smarts string
    """

    if smarts is None:
        return None
    return Chem.MolFromSmarts(smarts)  # type: ignore


def to_df(
    mols: List[Mol],
    smiles_column: Optional[str] = "smiles",
    mol_column: Optional[str] = None,
    include_private: bool = False,
    include_computed: bool = False,
    render_df_mol: bool = True,
    render_all_df_mol: bool = False,
    n_jobs: Optional[int] = 1,
) -> pd.DataFrame:
    """Convert a list of mols to a dataframe using each mol properties
    as a column.

    For the reverse operation, you might to check `dm.from_df()`.

    Args:
        mols: a molecule.
        smiles_column: name of the SMILES column.
        mol_column: Name of the column. If not None, rdkit.Chem.PandaTools
            is used to add a molecule column.
        include_private: Include private properties in the columns.
        include_computed: Include computed properties in the columns.
        render_df_mol: whether to render the molecule in the dataframe to images.
            If called once, it will be applied for the newly created dataframe with
            mol in it.
        render_all_df_mol: Whether to render all pandas dataframe mol column as images.
        n_jobs: Number of jobs for parallelization. Leave to 1 for no
            parallelization. Set to -1 to use all available cores.
    """

    # Init a dataframe
    df = pd.DataFrame()

    # Feed it with smiles
    if smiles_column is not None:
        smiles = dm.parallelized(to_smiles, mols, n_jobs=n_jobs)
        df[smiles_column] = smiles

    # Add a mol column
    if mol_column is not None:
        df[mol_column] = mols

    # Add any other properties present in the molecule
    def _mol_to_prop_dict(mol):
        if mol is not None:
            return mol.GetPropsAsDict(
                includePrivate=include_private,
                includeComputed=include_computed,
            )
        else:
            return {}

    # EN: You cannot use `processes` here because all properties will be lost
    # An alternative would be https://www.rdkit.org/docs/source/rdkit.Chem.PropertyMol.html
    # But this has less overhead
    props = dm.parallelized(_mol_to_prop_dict, mols, n_jobs=n_jobs, scheduler="threads")
    props_df = pd.DataFrame(props)
    if smiles_column is not None and smiles_column in props_df.columns:
        logger.warning(
            f"The SMILES column name provided ('{smiles_column}') is already present in the properties"
            " of the molecules. THe returned dataframe will two columns with the same name."
        )

    # Concat the df with the properties df
    df = pd.concat([df, props_df], axis=1)

    # Render mol column to images
    if render_df_mol is True and mol_column is not None:
        render_mol_df(df)

        if render_all_df_mol:
            PandasTools.RenderImagesInAllDataFrames()

    return df


def from_df(
    df: pd.DataFrame,
    smiles_column: Optional[str] = "smiles",
    mol_column: Optional[str] = None,
    conserve_smiles: bool = False,
    sanitize: bool = True,
) -> List[Mol]:
    """Convert a dataframe to a list of mols.

    For the reverse operation, you might to check `dm.to_df()`.

    Note:
        If `smiles_column` is used to build the molecules, this property
        is removed from the molecules' properties. You can decide to conserve
        the SMILES column by setting `conserve_smiles` to True.

    Args:
        df: a dataframe.
        smiles_column: Column name to extract the molecule.
        mol_column: Column name to extract the molecule. It takes
            precedence over `smiles_column`.
        conserve_smiles: Whether to conserve the SMILES in the mols' props.
        sanitize: Whether to sanitize if `smiles_column` is not None.
    """

    if smiles_column is None and mol_column is None:
        raise ValueError("Either `smiles_column` or `mol_column` must be not None.")

    if len(df) == 0:
        return []

    # Try to detect the mol column if `mol_column` is None.
    if mol_column is None:
        for col in df.columns:
            if isinstance(df[col].iloc[0], Mol):
                col = cast(str, col)
                mol_column = col

    def _row_to_mol(row) -> Optional[Mol]:
        props = row.to_dict()

        if mol_column is not None:
            mol = props.pop(mol_column)
        else:
            if conserve_smiles:
                smiles = props[smiles_column]
            else:
                # If a SMILES column is used to create the molecule then it is removed from the
                # properties.
                smiles = props.pop(smiles_column)

            mol = dm.to_mol(smiles, sanitize=sanitize)

        if mol is None:
            return None

        dm.set_mol_props(mol, props)
        return mol

    return df.apply(_row_to_mol, axis=1).tolist()  # type: ignore


def render_mol_df(df: pd.DataFrame):
    """Render the molecules column in a dataframe. The rendering is performed
    in-place only. So nothing is returned.

    Args:
        df: a dataframe.
    """

    # NOTE(hadim): _ChangeMoleculeRendering is not relevant anymore with rdkit>=2022.09
    if dm.is_lower_than_current_rdkit_version("2022.09"):
        # NOTE(hadim): replace by `PandaTools.ChangeMoleculeRendering` once
        # https://github.com/rdkit/rdkit/issues/3563 is fixed.
        _ChangeMoleculeRendering(df)


def _ChangeMoleculeRendering(frame=None, renderer="PNG"):
    """Allows to change the rendering of the molecules between base64 PNG images and string
    representations.
    This serves two purposes: First it allows to avoid the generation of images if this is
    not desired and, secondly, it allows to enable image rendering for newly created dataframe
    that already contains molecules, without having to rerun the time-consuming
    AddMoleculeColumnToFrame. Note: this behaviour is, because some pandas methods, e.g. head()
    returns a new dataframe instance that uses the default pandas rendering (thus not drawing
    images for molecules) instead of the monkey-patched one.
    """
    import types

    if renderer == "String":
        Chem.rdchem.Mol.__str__ = PandasTools.PrintDefaultMolRep  # type: ignore
    else:
        Chem.rdchem.Mol.__str__ = PandasTools.PrintAsBase64PNGString

    if frame is not None:
        frame.to_html = types.MethodType(PandasTools.patchPandasHTMLrepr, frame)  # type: ignore

    if PandasTools.defPandasRepr is not None and renderer == "String":  # type: ignore
        frame._repr_html_ = types.MethodType(PandasTools.defPandasRepr, frame)  # type: ignore
    else:
        frame._repr_html_ = types.MethodType(PandasTools.patchPandasrepr, frame)  # type: ignore


def _process_inchi_options(
    fixed_hydrogen_layer: bool = True,
    undefined_stereocenter: bool = True,
    reconnected_metal_layer: bool = True,
    tautomerism_keto_enol: bool = True,
    tautomerism_15: bool = True,
    options: Optional[List[str]] = None,
):
    inchi_options = []

    if fixed_hydrogen_layer:
        inchi_options.append("/FixedH")

    if undefined_stereocenter:
        inchi_options.append("/SUU")

    if reconnected_metal_layer:
        inchi_options.append("/RecMet")

    if tautomerism_keto_enol:
        inchi_options.append("/KET")

    if tautomerism_15:
        inchi_options.append("/15T")

    if options is not None:
        inchi_options.extend(options)

    inchi_options = " ".join(inchi_options)
    return inchi_options



import pytest

import pandas as pd
import datamol as dm


def test_descriptors():
    smiles_list = ["CC(=O)OC1=CC=CC=C1C(=O)O", "CCN(CC)CCCC(C)NC1=C2C=CC(=CC2=NC=C1)Cl"]

    for smiles in smiles_list:
        mol = dm.to_mol(smiles)

        dm.descriptors.mw(mol)
        dm.descriptors.fsp3(mol)
        dm.descriptors.n_hba(mol)
        dm.descriptors.n_hbd(mol)
        dm.descriptors.n_lipinski_hba(mol)
        dm.descriptors.n_lipinski_hbd(mol)
        dm.descriptors.n_rings(mol)
        dm.descriptors.n_hetero_atoms(mol)
        dm.descriptors.n_heavy_atoms(mol)
        dm.descriptors.n_rotatable_bonds(mol)
        dm.descriptors.n_aliphatic_rings(mol)
        dm.descriptors.n_aromatic_rings(mol)
        dm.descriptors.n_saturated_rings(mol)
        dm.descriptors.n_radical_electrons(mol)
        dm.descriptors.tpsa(mol)
        dm.descriptors.qed(mol)
        dm.descriptors.clogp(mol)
        dm.descriptors.sas(mol)
        dm.descriptors.sas(mol)
        dm.descriptors.n_stereo_centers_unspecified(mol)
        dm.descriptors.n_spiro_atoms(mol)

        dm.descriptors.n_aliphatic_carbocycles(mol)
        dm.descriptors.n_aliphatic_heterocyles(mol)
        dm.descriptors.n_aliphatic_rings(mol)
        dm.descriptors.n_aromatic_carbocycles(mol)
        dm.descriptors.n_aromatic_heterocyles(mol)
        dm.descriptors.n_aromatic_rings(mol)
        dm.descriptors.n_saturated_carbocycles(mol)
        dm.descriptors.n_saturated_heterocyles(mol)
        dm.descriptors.n_saturated_rings(mol)


def test_compute_many_descriptors():
    mol = dm.to_mol("CCN(CC)CCCC(C)NC1=C2C=CC(=CC2=NC=C1)Cl")

    true_values = pd.Series(
        {
            "mw": 319.181525512,
            "fsp3": 0.5,
            "n_lipinski_hba": 3.0,
            "n_lipinski_hbd": 1.0,
            "n_rings": 2.0,
            "n_hetero_atoms": 4.0,
            "n_heavy_atoms": 22.0,
            "n_rotatable_bonds": 8.0,
            "n_radical_electrons": 0.0,
            "tpsa": 28.16,
            "qed": 0.7564117572128701,
            "clogp": 4.810600000000004,
            "sas": 2.670786229594949,
            "n_aliphatic_carbocycles": 0.0,
            "n_aliphatic_heterocyles": 0.0,
            "n_aliphatic_rings": 0.0,
            "n_aromatic_carbocycles": 1.0,
            "n_aromatic_heterocyles": 1.0,
            "n_aromatic_rings": 2.0,
            "n_saturated_carbocycles": 0.0,
            "n_saturated_heterocyles": 0.0,
            "n_saturated_rings": 0.0,
        }
    )

    # Scenario #1
    props = dm.descriptors.compute_many_descriptors(mol)
    props = pd.Series(props)

    assert props.equals(true_values)

    # Scenario #2
    props = dm.descriptors.compute_many_descriptors(
        mol,
        properties_fn={"hello": lambda x: 88},
        add_properties=False,
    )
    assert props == {"hello": 88}

    # Scenario #3
    props = dm.descriptors.compute_many_descriptors(
        mol,
        properties_fn={"hello": lambda x: 88},
        add_properties=True,
    )
    props = pd.Series(props)

    true_values_2 = true_values.copy()
    true_values_2["hello"] = 88
    true_values_2 = true_values_2[props.index]

    assert true_values_2.equals(props)


def test_compute_many_descriptors_with_function_as_string():
    mol = dm.to_mol("CC(=O)OC1=CC=CC=C1C(=O)O")

    results = dm.descriptors.compute_many_descriptors(
        mol,
        properties_fn={"max_partial_charge": "MaxPartialCharge"},
        add_properties=False,
    )

    assert "max_partial_charge" in results.keys()
    assert pytest.approx(0.33900378687731025) == results["max_partial_charge"]


def test_batch_compute_many_descriptors():
    data = dm.data.freesolv()
    data = data.iloc[:30]
    mols = data["smiles"].apply(dm.to_mol).tolist()

    props = dm.descriptors.batch_compute_many_descriptors(
        mols,
        batch_size=64,
        n_jobs=-1,
        progress=False,
    )

    assert set(props.columns.tolist()) == {
        "mw",
        "fsp3",
        "n_lipinski_hba",
        "n_lipinski_hbd",
        "n_rings",
        "n_hetero_atoms",
        "n_heavy_atoms",
        "n_rotatable_bonds",
        "n_radical_electrons",
        "tpsa",
        "qed",
        "clogp",
        "sas",
        "n_aliphatic_carbocycles",
        "n_aliphatic_heterocyles",
        "n_aliphatic_rings",
        "n_aromatic_carbocycles",
        "n_aromatic_heterocyles",
        "n_aromatic_rings",
        "n_saturated_carbocycles",
        "n_saturated_heterocyles",
        "n_saturated_rings",
    }
    assert props.shape == (30, 22)


def test_any_rdkit_descriptor():
    mol = dm.to_mol("CC(=O)OC1=CC=CC=C1C(=O)O")

    value = dm.descriptors.any_rdkit_descriptor("MaxPartialCharge")(mol)
    assert pytest.approx(value) == 0.33900378687731025

    value = dm.descriptors.any_rdkit_descriptor("CalcFractionCSP3")(mol)
    assert pytest.approx(value) == 0.1111111111111111

    with pytest.raises(ValueError):
        dm.descriptors.any_rdkit_descriptor("DOES NOT EXIST")


def test_n_aromatic_atoms():
    smiles = "Nc1cnn(-c2ccccc2)c(=O)c1Cl"
    mol = dm.to_mol(smiles)

    assert dm.descriptors.n_aromatic_atoms(mol) == 12
    assert dm.descriptors.n_aromatic_atoms_proportion(mol) == 0.8


def test_formal_charge():
    mol = dm.to_mol("CC(=O)NC1CCC2=CC(=C(C(=C2C3=CC=C(C(=O)C=C13)OC)OC)OC)OC")
    assert dm.descriptors.formal_charge(mol) == 0

    mol = dm.to_mol("C(CC(=O)[O-])C(C(=O)[O-])[NH3+]")
    assert dm.descriptors.formal_charge(mol) == -1


def test_refractivity():
    mol = dm.to_mol("CN1C(=O)CN=C(C2=C1C=CC(=C2)Cl)C3=CC=CC=C3")

    value = dm.descriptors.refractivity(mol)
    assert pytest.approx(value, rel=2) == 81.10


def test_n_rigid_bonds():
    mol = dm.to_mol("CC(=O)NC1CCC2=CC(=C(C(=C2C3=CC=C(C(=O)C=C13)OC)OC)OC)OC")
    assert dm.descriptors.n_rigid_bonds(mol) == 20

    mol = dm.to_mol("CN1C(=O)CN=C(C2=C1C=CC(=C2)Cl)C3=CC=CC=C3")
    assert dm.descriptors.n_rigid_bonds(mol) == 19


def test_n_stereocenters():
    mol = dm.to_mol("CC(=O)NC1CCC2=CC(=C(C(=C2C3=CC=C(C(=O)C=C13)OC)OC)OC)OC")

    assert dm.descriptors.n_stereo_centers(mol) == 1

    mol = dm.to_mol("CN1C(=O)CN=C(C2=C1C=CC(=C2)Cl)C3=CC=CC=C3")
    assert dm.descriptors.n_stereo_centers(mol) == 0


def test_n_charged_atoms():
    mol = dm.to_mol("C(CC(=O)[O-])C(C(=O)[O-])[NH3+]")
    assert dm.descriptors.n_charged_atoms(mol) == 3



from typing import List
from typing import Any

from rdkit.Chem import rdFMCS

import datamol as dm

ALLOWED_ATOM_COMPARE = ["CompareAny", "CompareAnyHeavyAtom", "CompareElements", "CompareIsotopes"]
ALLOWED_BOND_COMPARE = ["CompareAny", "CompareOrder", "CompareOrderExact"]
ALLOWED_RING_COMPARE = ["IgnoreRingFusion", "PermissiveRingFusion", "StrictRingFusion"]


def find_mcs(
    mols: List[dm.Mol],
    maximize_bonds: bool = True,
    threshold: float = 0.0,
    timeout: int = 5,
    verbose: bool = False,
    match_valences: bool = False,
    ring_matches_ring_only: bool = True,
    complete_rings_only: bool = False,
    match_chiral_tag: bool = False,
    seed_smarts: str = "",
    atom_compare: str = "CompareElements",
    bond_compare: str = "CompareOrder",
    ring_compare: str = "IgnoreRingFusion",
    with_details: bool = False,
    **kwargs: Any,
):
    """Find the maximum common substructure from a list of molecules.

    Args:
        mols: List of molecules.
        maximize_bonds: Maximize the number of bonds in the substructure.
        threshold: The threshold for the MCS (between 0 and 1).
        timeout: The timeout for the MCS.
        verbose: Whether to enable verbose mode.
        match_valences: Whether to match valences.
        ring_matches_ring_only: Whether to match rings only.
        complete_rings_only: Whether to match complete rings only.
        match_chiral_tag: Whether to match chiral tags.
        seed_smarts: The seed SMARTS.
        atom_compare: One of "CompareAny", "CompareAnyHeavyAtom", "CompareElements",
            "CompareIsotopes".
        bond_compare: One of "CompareAny", "CompareOrder", "CompareOrderExact".
        ring_compare: One of "IgnoreRingFusion", "PermissiveRingFusion", "StrictRingFusion".
        with_details: Whether to return the RDKit MCS object or just the SMARTS string.
        **kwargs: Additional arguments for the MCS.
    """

    if atom_compare not in ALLOWED_ATOM_COMPARE:
        raise ValueError(f"atom_compare must be one of {ALLOWED_ATOM_COMPARE}")

    if bond_compare not in ALLOWED_BOND_COMPARE:
        raise ValueError(f"bond_compare must be one of {ALLOWED_BOND_COMPARE}")

    if ring_compare not in ALLOWED_RING_COMPARE:
        raise ValueError(f"ring_compare must be one of {ALLOWED_RING_COMPARE}")

    args = {}
    args["maximizeBonds"] = maximize_bonds
    args["threshold"] = threshold
    args["timeout"] = timeout
    args["verbose"] = verbose
    args["matchValences"] = match_valences
    args["ringMatchesRingOnly"] = ring_matches_ring_only
    args["completeRingsOnly"] = complete_rings_only
    args["matchChiralTag"] = match_chiral_tag
    args["seedSmarts"] = seed_smarts
    args["atomCompare"] = rdFMCS.AtomCompare.names[atom_compare]
    args["bondCompare"] = rdFMCS.BondCompare.names[bond_compare]
    args["ringCompare"] = rdFMCS.RingCompare.names[ring_compare]

    args.update(kwargs)

    mcs = rdFMCS.FindMCS(mols, **args)

    if with_details:
        return mcs

    smarts = mcs.smartsString
    if smarts == "":
        smarts = None
    return smarts



import warnings

import pytest

import datamol as dm


def test_to_fp():
    smiles = "CC(=O)Oc1ccccc1C(=O)O"
    mol = dm.to_mol(smiles)

    assert dm.to_fp(mol).shape[0] == 2048
    assert dm.to_fp(mol).sum() == 31


def test_list_fp():
    assert set(dm.list_supported_fingerprints().keys()) == {
        "atompair",
        "atompair-count",
        "avalon-count",
        "ecfp",
        "fcfp",
        "ecfp-count",
        "erg",
        "estate",
        "fcfp-count",
        "layered",
        "maccs",
        "pattern",
        "rdkit",
        "topological",
        "topological-count",
        "rdkit-count",
    }


def test_all_fps():
    smiles = "CC(=O)Oc1ccccc1C(=O)O"
    mol = dm.to_mol(smiles)

    fp_infos = {}
    for fp_type in dm.list_supported_fingerprints():
        fold_size = None
        if fp_type == "rdkit-count":
            fold_size = 2048

        print(fp_type)
        args = {}
        args["mol"] = mol
        args["as_array"] = True
        args["fp_type"] = fp_type
        args["fold_size"] = fold_size
        fp = dm.to_fp(**args)

        fp_infos[fp_type] = dict(size=len(fp), bits_sum=fp.sum())

    print(fp_infos)

    assert fp_infos == {
        "maccs": {"size": 167, "bits_sum": 21},
        "ecfp": {"size": 2048, "bits_sum": 31},
        "fcfp": {"size": 2048, "bits_sum": 22},
        "topological": {"size": 2048, "bits_sum": 18},
        "atompair": {"size": 2048, "bits_sum": 68},
        "rdkit": {"size": 2048, "bits_sum": 354},
        "pattern": {"size": 2048, "bits_sum": 173},
        "layered": {"size": 2048, "bits_sum": 335},
        "erg": {"size": 315, "bits_sum": 23.4},
        "estate": {"size": 79, "bits_sum": 13},
        "avalon-count": {"size": 512, "bits_sum": 168},
        "ecfp-count": {"size": 2048, "bits_sum": 35},
        "fcfp-count": {"size": 2048, "bits_sum": 35},
        "topological-count": {"size": 2048, "bits_sum": 19},
        "atompair-count": {"size": 2048, "bits_sum": 78},
        "rdkit-count": {"size": 2048, "bits_sum": 301},
    }


def test_fp_invalid_input():
    args = {}
    args["mol"] = None
    args["radius"] = 3

    with pytest.raises(ValueError):
        dm.to_fp(**args)

    args["mol"] = "dsdsdsd"
    with pytest.raises(ValueError):
        dm.to_fp(**args)



import pytest
import time

from rdkit import Chem

import datamol as dm


def test_enumerate_tautomers():
    mol = dm.to_mol("OC1=CC2CCCCC2[N:1]=C1")

    mols = dm.enumerate_tautomers(mol, n_variants=10)

    assert {dm.to_smiles(m) for m in mols} == {"O=C1C=[N:1]C2CCCCC2C1", "OC1=CC2CCCCC2[N:1]=C1"}


def test_enumerate_stereo():
    mol = dm.to_mol("OC1=CC2CCCCC2[N:1]=C1")

    mols = dm.enumerate_stereoisomers(mol, n_variants=10)

    assert {dm.to_smiles(m) for m in mols} == {
        "OC1=C[C@@H]2CCCC[C@@H]2[N:1]=C1",
        "OC1=C[C@@H]2CCCC[C@H]2[N:1]=C1",
        "OC1=C[C@H]2CCCC[C@@H]2[N:1]=C1",
        "OC1=C[C@H]2CCCC[C@H]2[N:1]=C1",
    }


def test_enumerate_stereo_undefined_failure():
    mol = dm.to_mol(
        "N=1C(NC2CC2)=C3C(=NC1)N(/C=C/C=4C=C(C=CC4C)C(=O)NC=5C=C(C=C(C5)N6CCN(CC6)C)C(F)(F)F)C=N3"
    )
    with pytest.raises(RuntimeError):
        dm.enumerate_stereoisomers(mol, clean_it=True)

    mols = dm.enumerate_stereoisomers(mol, clean_it=False)
    assert len(mols) == 2  # only one double bond


def test_enumerate_stereo_timeout():
    mol = dm.to_mol(
        "CC1=C2C(C(=O)C3(C(CC4C(C3C(C(C2(C)C)(CC1OC(=O)C(C(C5=CC=CC=C5)NC(=O)C6=CC=CC=C6)O)O)OC(=O)C7=CC=CC=C7)(CO4)OC(=O)C)O)C)OC(=O)C"
    )

    # NOTE(hadim): it's impossible to predict anything given a timeout for different
    # machines so we here we just check the code can run without errors
    dm.enumerate_stereoisomers(mol, n_variants=10, timeout_seconds=1)


def test_enumerate_structural():
    mol = dm.to_mol("CCCCC")  # pentane has only three structural isomers

    mols_iso = dm.enumerate_structisomers(
        mol,
        n_variants=5,
        allow_cycle=False,
        depth=2,
        allow_double_bond=False,
        allow_triple_bond=False,
    )

    assert {dm.to_smiles(m) for m in mols_iso} == {"CCC(C)C", "CC(C)(C)C"}

    # NOTE(hadim): disable to reduce testing time
    # mols_cyclo_iso = dm.enumerate_structisomers(mol, n_variants=5, depth=2, allow_cycle=True)

    # # expect 3 molecules with cycles
    # assert sum([Chem.rdMolDescriptors.CalcNumRings(x) == 1 for x in mols_cyclo_iso]) == 3  # type: ignore

    # mols_cyclo_iso_double = dm.enumerate_structisomers(
    #     mol, n_variants=10, allow_cycle=True, allow_double_bond=True
    # )
    # should have mol with double link
    # assert sum(["=" in dm.to_smiles(x) for x in mols_cyclo_iso_double]) > 0


@pytest.mark.skip_platform("win")
def test_enumerate_structural_timeout():
    mol = dm.to_mol("CCCCC")

    # NOTE(hadim): it's impossible to predict anything given a timeout for different
    # machines so we here we just check the code can run without errors
    dm.enumerate_structisomers(mol, n_variants=10, timeout_seconds=1)


def test_canonical_tautomer():
    smiles = "Oc1c(cccc3)c3nc2ccncc12"
    mol = dm.to_mol(smiles)

    canonical_mol = dm.canonical_tautomer(mol)

    assert dm.to_smiles(canonical_mol) == "O=c1c2ccccc2[nH]c2ccncc12"
    assert dm.to_inchikey(canonical_mol) == dm.to_inchikey(mol)


def test_remove_stereochemistry():
    mol = dm.to_mol("C[C@H]1CCC[C@@H](C)[C@@H]1Cl")
    mol_no_stereo = dm.remove_stereochemistry(mol)
    assert dm.to_smiles(mol_no_stereo) == "CC1CCCC(C)C1Cl"



from ._conformers import generate
from ._conformers import cluster
from ._conformers import rmsd
from ._conformers import return_centroids
from ._conformers import translate
from ._conformers import align_conformers

from ._features import sasa
from ._features import get_coords
from ._features import center_of_mass
from ._features import keep_conformers



import datamol as dm


def test_watch_duration():
    def fn(n):
        for i in range(n):
            print(i)

    with dm.utils.perf.watch_duration(log=True) as w:
        fn(5)

    assert isinstance(w.duration, float)



"""
The data module aims to provide a fast and convenient access to various molecular datasets.

---
"""

from typing import Optional
from typing import cast
from typing import Union
from typing import List
from typing import overload
from typing import Literal

import pkg_resources

import pandas as pd

from .types import Mol
from .io import read_sdf
from .convert import from_df
from .convert import render_mol_df


@overload
def freesolv(as_df: Literal[True] = True) -> pd.DataFrame:
    ...


@overload
def freesolv(as_df: Literal[False] = False) -> List[Mol]:
    ...


@overload
def freesolv(as_df: bool = True) -> Union[List[Mol], pd.DataFrame]:
    ...


def freesolv(as_df: bool = True) -> Union[List[Mol], pd.DataFrame]:
    """Return the FreeSolv dataset as a dataframe.

    The dataset contains 642 molecules and the following columns:
    `['iupac', 'smiles', 'expt', 'calc']`.

    Warning:
        This dataset is only meant to be used as a toy dataset for pedagogic and
        testing purposes. **It is not** a dataset for benchmarking, analysis or
        model training.
    """

    with pkg_resources.resource_stream("datamol", "data/freesolv.csv") as f:
        data = pd.read_csv(f)

    if not as_df:
        data = from_df(data)

    return data


@overload
def cdk2(as_df: Literal[True] = True, mol_column: Optional[str] = "mol") -> pd.DataFrame:
    ...


@overload
def cdk2(as_df: Literal[False] = False, mol_column: Optional[str] = "mol") -> List[Mol]:
    ...


@overload
def cdk2(as_df: bool = True, mol_column: Optional[str] = "mol") -> Union[List[Mol], pd.DataFrame]:
    ...


def cdk2(as_df: bool = True, mol_column: Optional[str] = "mol"):
    """Return the RDKit CDK2 dataset from `RDConfig.RDDocsDir, 'Book/data/cdk2.sdf'`.

    Args:
        as_df: Whether to return a list mol or a pandas DataFrame.
        mol_column: Name of the mol column. Only relevant if `as_df` is True.
    """

    with pkg_resources.resource_stream("datamol", "data/cdk2.sdf") as f:
        data = read_sdf(f, as_df=as_df, mol_column=mol_column)
    return data


@overload
def solubility(as_df: Literal[True] = True, mol_column: Optional[str] = "mol") -> pd.DataFrame:
    ...


@overload
def solubility(as_df: Literal[False] = False, mol_column: Optional[str] = "mol") -> List[Mol]:
    ...


@overload
def solubility(
    as_df: bool = True, mol_column: Optional[str] = "mol"
) -> Union[List[Mol], pd.DataFrame]:
    ...


def solubility(as_df: bool = True, mol_column: Optional[str] = "mol"):
    """Return the RDKit solubility dataset from `RDConfig.RDDocsDir, 'Book/data/solubility.{train|test}.sdf'`.

    The dataframe or the list of molecules with contain a `split` column, either `train` or `test`.

    Args:
        as_df: Whether to return a list mol or a pandas DataFrame.
        mol_column: Name of the mol column. Only relevant if `as_df` is True.
    """

    with pkg_resources.resource_stream("datamol", "data/solubility.train.sdf") as f:
        train = read_sdf(f, as_df=True, mol_column="mol", smiles_column=None)

    with pkg_resources.resource_stream("datamol", "data/solubility.test.sdf") as f:
        test = read_sdf(f, as_df=True, mol_column="mol", smiles_column=None)

    train = cast(pd.DataFrame, train)
    test = cast(pd.DataFrame, test)

    train["split"] = "train"
    test["split"] = "test"

    # NOTE(hadim): LMAO RDkit consistency xD
    test = test.rename(columns={"SMILES": "smiles"})

    data = pd.concat([train, test], ignore_index=True)

    if as_df:
        if mol_column is None:
            data = data.drop(columns=["mol"])

        render_mol_df(data)
        return data

    return from_df(data, mol_column=mol_column)



import time

from rdkit import Chem
from rdkit.Chem import rdmolops

from rdkit.Chem.MolStandardize import rdMolStandardize
from rdkit.Chem.EnumerateStereoisomers import EnumerateStereoisomers
from rdkit.Chem.EnumerateStereoisomers import StereoEnumerationOptions

import datamol as dm

from ._structural import IsomerEnumerator


def enumerate_tautomers(
    mol: dm.Mol,
    n_variants: int = 20,
    max_transforms: int = 1000,
    reassign_stereo: bool = True,
    remove_bond_stereo: bool = True,
    remove_sp3_stereo: bool = True,
):
    """Enumerate the possible tautomers of the current molecule.

    Args:
        mol: The molecule whose state we should enumerate.
        n_variants: The maximum amount of molecules that should be returned.
        max_transforms: Set the maximum number of transformations to be applied. This limit is usually
            hit earlier than the n_variants limit and leads to a more linear scaling
            of CPU time with increasing number of tautomeric centers (see Sitzmann et al.).
        reassign_stereo: Whether to reassign stereo centers.
        remove_bond_stereo: Whether stereochemistry information is removed from double bonds involved in tautomerism.
            This means that enols will lose their E/Z stereochemistry after going through tautomer enumeration because
            of the keto-enolic tautomerism.
        remove_sp3_stereo: Whether stereochemistry information is removed from sp3 atoms involved in tautomerism. This
            means that S-aminoacids will lose their stereochemistry after going through tautomer enumeration because
            of the amido-imidol tautomerism.
    """
    enumerator = rdMolStandardize.TautomerEnumerator()

    # Configure the enumeration
    enumerator.SetMaxTautomers(n_variants)
    enumerator.SetMaxTransforms(max_transforms)
    enumerator.SetReassignStereo(reassign_stereo)
    enumerator.SetRemoveBondStereo(remove_bond_stereo)
    enumerator.SetRemoveSp3Stereo(remove_sp3_stereo)

    tautomers = enumerator.Enumerate(mol)
    return list(tautomers)


def enumerate_stereoisomers(
    mol: dm.Mol,
    n_variants: int = 20,
    undefined_only: bool = False,
    rationalise: bool = True,
    timeout_seconds: int = None,
    clean_it: bool = True,
):
    """Enumerate the stereocenters and bonds of the current molecule.

    Original source: the `openff-toolkit` lib.

    Warning: this function can be computationnaly intensive.

    Args:
        mol: The molecule whose state we should enumerate.
        n_variants: The maximum amount of molecules that should be returned.
        undefined_only: If we should enumerate all stereocenters and bonds or only those
            with undefined stereochemistry.
        rationalise: If we should try to build and rationalise the molecule to ensure it
            can exist.
        timeout_seconds: The maximum amount of time to spend on enumeration. None
            will disable the timeout. Note that the timeout might be inaccurate as a running single variant
            computation is not stopped when the duration is reached.
        clean_it: A flag for assigning stereochemistry. If True, it will remove previous stereochemistry
            markings on the bonds.
    """

    # safety first
    mol = dm.copy_mol(mol)

    # in case any bonds/centers are missing stereo chem flag it here
    Chem.AssignStereochemistry(mol, force=False, flagPossibleStereoCenters=True, cleanIt=clean_it)  # type: ignore
    Chem.FindPotentialStereoBonds(mol, cleanIt=clean_it)  # type: ignore

    # set up the options
    stereo_opts = StereoEnumerationOptions(
        tryEmbedding=rationalise,
        onlyUnassigned=undefined_only,
        maxIsomers=n_variants,
        unique=True,
    )

    isomers_iterator = EnumerateStereoisomers(mol, options=stereo_opts)

    start = time.time()
    duration = 0

    isomers = []
    while timeout_seconds is None or duration < timeout_seconds:
        try:
            isomer = next(isomers_iterator)
            isomers.append(isomer)
        except StopIteration:
            break

        duration = time.time() - start

    variants = []
    for isomer in isomers:
        # isomer has CIS/TRANS tags so convert back to E/Z
        Chem.SetDoubleBondNeighborDirections(isomer)  # type: ignore
        Chem.AssignStereochemistry(isomer, force=True, cleanIt=clean_it)  # type: ignore
        variants.append(isomer)

    return variants


def enumerate_structisomers(
    mol: dm.Mol,
    n_variants: int = 20,
    allow_cycle: bool = False,
    allow_double_bond: bool = False,
    allow_triple_bond: bool = False,
    depth: int = None,
    timeout_seconds: int = None,
):
    """Enumerate the structural isomers of the input molecule

    Warning: this function can be computationnaly intensive.

    Args:
        mol: The molecule whose state we should enumerate.
        n_variants: The maximum amount of molecules that should be returned.
        allow_cycle: whether to allow transformation involving cycle creation.
        allow_double_bond: whether to allow transformation involving at least one double bond.
        allow_triple_bond: whether to allow transformation involving at least one triple bond.
        depth: depth of the search, use a sensible value to decrease search time. Will fall back to number of atoms.
        timeout_seconds: The maximum amount of time to spend on enumeration. None
            will disable the timeout. Note that the timeout might be inaccurate as a running single variant
            computation is not stopped when the duration is reached.
    """

    mol = dm.copy_mol(mol)

    enumerator = IsomerEnumerator(
        allow_cycle=allow_cycle,
        allow_double_bond=allow_double_bond,
        allow_triple_bond=allow_triple_bond,
    )

    if depth is None:
        depth = mol.GetNumAtoms()

    isomers_iterator = enumerator(
        mol,
        max_mols=n_variants,
        include_input=False,
        depth=depth,
    )

    start = time.time()
    duration = 0

    isomers = []
    while timeout_seconds is None or duration < timeout_seconds:
        try:
            isomer = next(isomers_iterator)
            isomers.append(dm.to_mol(isomer))
        except StopIteration:
            break

        duration = time.time() - start

    return isomers


def remove_stereochemistry(mol: dm.Mol, copy: bool = True):
    """Removes all stereochemistry info from the molecule.

    Args:
        mol: a molecule
        copy: Whether to copy the molecule.
    """

    if copy is True:
        mol = dm.copy_mol(mol)
    rdmolops.RemoveStereochemistry(mol)
    return mol


def canonical_tautomer(mol: dm.Mol):
    """Get the canonical tautomer of the current molecule.

    Args:
        mol: A molecule.
    """
    enumerator = rdMolStandardize.TautomerEnumerator()
    return enumerator.Canonicalize(mol)



from .descriptors import mw
from .descriptors import fsp3
from .descriptors import n_hba
from .descriptors import n_hbd
from .descriptors import n_lipinski_hba
from .descriptors import n_lipinski_hbd
from .descriptors import n_rings
from .descriptors import n_hetero_atoms
from .descriptors import n_heavy_atoms
from .descriptors import n_rotatable_bonds
from .descriptors import n_radical_electrons
from .descriptors import tpsa
from .descriptors import qed
from .descriptors import clogp
from .descriptors import sas
from .descriptors import n_NHOH
from .descriptors import n_NO
from .descriptors import formal_charge
from .descriptors import n_aliphatic_carbocycles
from .descriptors import n_aliphatic_heterocyles
from .descriptors import n_aliphatic_rings
from .descriptors import n_aromatic_carbocycles
from .descriptors import n_aromatic_heterocyles
from .descriptors import n_aromatic_rings
from .descriptors import n_saturated_carbocycles
from .descriptors import n_saturated_heterocyles
from .descriptors import n_saturated_rings
from .descriptors import n_aromatic_atoms
from .descriptors import n_aromatic_atoms_proportion
from .descriptors import refractivity
from .descriptors import n_rigid_bonds
from .descriptors import n_stereo_centers
from .descriptors import n_charged_atoms
from .descriptors import n_stereo_centers_unspecified
from .descriptors import n_spiro_atoms

from .compute import any_rdkit_descriptor
from .compute import compute_many_descriptors
from .compute import batch_compute_many_descriptors



from typing import cast
import pytest

import numpy as np
import pandas as pd
from rdkit import Chem
from selfies import __version__ as selfies_version
import datamol as dm


def test_to_smiles():
    smiles = "O=C(C)Oc1ccccc1C(=O)O"
    mol = dm.to_mol(smiles)

    smiles = dm.to_smiles(
        mol,
        isomeric=True,
        ordered=True,
        explicit_bonds=False,
        explicit_hs=False,
    )
    assert smiles == "CC(=O)Oc1ccccc1C(=O)O"

    smiles = dm.to_smiles(
        mol,
        isomeric=True,
        ordered=False,
        explicit_bonds=True,
        explicit_hs=False,
    )
    assert smiles == "C-C(=O)-O-c1:c:c:c:c:c:1-C(=O)-O"

    smiles = dm.to_smiles(
        mol,
        isomeric=True,
        ordered=False,
        explicit_bonds=False,
        explicit_hs=True,
    )
    assert smiles == "[CH3][C](=[O])[O][c]1[cH][cH][cH][cH][c]1[C](=[O])[OH]"

    mol = dm.to_mol(mol, kekulize=True)
    smiles = dm.to_smiles(
        mol, isomeric=True, ordered=False, explicit_bonds=False, explicit_hs=False, kekulize=True
    )
    assert smiles == "CC(=O)OC1=CC=CC=C1C(=O)O"

    smiles = "O=C(C)Oc1ccccc1C(=O)O"
    mol = dm.to_mol(smiles)
    randomized_smiles = dm.to_smiles(mol, randomize=True)
    randomized_mol = dm.to_mol(randomized_smiles)

    assert dm.to_smiles(randomized_mol) == dm.to_smiles(mol)


def test_to_selfies():
    smiles = "CC(=O)Oc1ccccc1C(=O)O"
    mol = dm.to_mol(smiles)

    true_sf = (
        "[C][C][=Branch1][C][=O][O][C][=C][C][=C][C][=C][Ring1][=Branch1][C][=Branch1][C][=O][O]"
    )
    if selfies_version < "2.0.0":
        true_sf = "[C][C][Branch1_2][C][=O][O][C][=C][C][=C][C][=C][Ring1][Branch1_2][C][Branch1_2][C][=O][O]"

    selfies = dm.to_selfies(smiles)
    assert selfies == true_sf

    selfies = dm.to_selfies(mol)
    assert selfies == true_sf


def test_from_selfies():
    # work with both selfies version
    selfies = (
        "[C][C][=Branch1][C][=O][O][C][=C][C][=C][C][=C][Ring1][=Branch1][C][=Branch1][C][=O][O]"
    )
    if selfies_version < "2.0.0":
        selfies = "[C][C][Branch1_2][C][=O][O][C][=C][C][=C][C][=C][Ring1][Branch1_2][C][Branch1_2][C][=O][O]"

    smiles = dm.from_selfies(selfies, as_mol=False)
    assert smiles == "CC(=O)OC1=CC=CC=C1C(=O)O"

    mol = dm.from_selfies(selfies, as_mol=True)
    assert dm.to_smiles(mol) == "CC(=O)Oc1ccccc1C(=O)O"


def test_smiles_as_smarts():
    smiles = "O=C(C)Oc1ccccc1C(=O)O"
    mol = dm.to_mol(smiles)

    smarts = dm.smiles_as_smarts(mol, keep_hs=True)
    assert smarts == "[CH3]-[C](=[O])-[O]-[c]1:[cH]:[cH]:[cH]:[cH]:[c]:1-[C](=[O])-[OH]"

    smarts = dm.smiles_as_smarts(smiles, keep_hs=False)
    assert smarts == "[CH3]-[C](=[O])-[O]-[c]1:[cH]:[cH]:[cH]:[cH]:[c]:1-[C](=[O])-[OH]"

    assert dm.smiles_as_smarts(None) is None


def test_to_smarts():
    smarts = "[OX2H][CX3]=[OX1]"
    mol = Chem.MolFromSmarts(smarts)
    expected_out = dm.to_smarts(mol)
    assert expected_out == "[O&X2&H1][C&X3]=[O&X1]"


def test_from_smarts():
    smarts = "[OX2H][CX3]=[OX1]"
    smiles = "O=C(C)Oc1ccccc1C(=O)O"
    query = dm.from_smarts(smarts)
    mol = dm.to_mol(smiles)
    assert mol.HasSubstructMatch(query)


def test_inchi():
    smiles = "CC(=O)Oc1ccccc1C(=O)O"
    mol = dm.to_mol(smiles)

    inchi = dm.to_inchi(mol)
    assert inchi == "InChI=1S/C9H8O4/c1-6(10)13-8-5-3-2-4-7(8)9(11)12/h2-5H,1H3,(H,11,12)"

    inchikey = dm.to_inchikey(mol)
    assert inchikey == "BSYNRYMUTXBXSQ-UHFFFAOYSA-N"

    new_mol = dm.from_inchi(inchi)
    assert dm.to_smiles(new_mol) == smiles

    assert dm.to_inchi(None) is None
    assert dm.to_inchi("") is None
    assert dm.to_inchi("C(C)(C)(C)(C)(C)") is None
    assert dm.to_inchikey(None) is None
    assert dm.to_inchikey("") is None
    assert dm.to_inchikey("C(C)(C)(C)(C)(C)") is None
    assert dm.from_inchi(None) is None


def test_to_df(datadir):
    data_path = datadir / "TUBB3-observations.sdf"
    mols = dm.read_sdf(data_path)
    df = dm.to_df(mols)

    assert df.shape == (10, 12)
    assert list(df.columns) == [
        "smiles",
        "zinc_id",
        "ortholog_name",
        "gene_name",
        "affinity",
        "chembldocid",
        "title",
        "reference.pubmed_id",
        "reference.doi",
        "reference.chembl_id",
        "reference.journal",
        "reference.year",
    ]

    # EN: more than 500 will slow the test
    large_mol_set = np.random.choice(mols, 500)
    with dm.utils.perf.watch_duration(log=True) as w:
        df_sequential = dm.to_df(large_mol_set, n_jobs=1)
    sequential_time = w.duration

    with w:
        df_parallel = dm.to_df(large_mol_set, n_jobs=-1)
    parallel_time = w.duration

    pd.testing.assert_frame_equal(df_sequential, df_parallel)
    # assert parallel_time < sequential_time


def test_from_df(datadir):
    data_path = datadir / "TUBB3-observations.sdf"
    df = dm.read_sdf(data_path, as_df=True)

    mols = dm.from_df(df)

    assert len(mols) == 10
    assert isinstance(mols[0], Chem.rdchem.Mol)

    assert set(mols[0].GetPropsAsDict().keys()) == {
        "zinc_id",
        "ortholog_name",
        "gene_name",
        "affinity",
        "chembldocid",
        "title",
        "reference.pubmed_id",
        "reference.doi",
        "reference.chembl_id",
        "reference.journal",
        "reference.year",
    }

    assert dm.from_df(pd.DataFrame()) == []


def test_from_df_conserve_smiles(datadir):
    data_path = datadir / "freesolv.csv"
    df = dm.read_csv(data_path)
    mols = dm.from_df(df, conserve_smiles=True)
    assert "smiles" in mols[0].GetPropsAsDict().keys()


def test_to_df_smiles_warning(datadir, caplog):
    data_path = datadir / "freesolv.csv"
    df = dm.read_csv(data_path)

    mols = dm.from_df(df, conserve_smiles=True)
    df = dm.to_df(mols)

    assert sum(df.columns == "smiles") == 2

    assert "WARNING" in caplog.text
    assert (
        "The SMILES column name provided ('smiles') is already present in the properties of the molecules"
        in caplog.text
    )


def test_to_cxsmiles():
    mol = dm.to_mol("OC1=CC2CCCCC2[N:1]=C1")
    smiles = dm.to_smiles(mol, cxsmiles=True)
    assert smiles == "OC1=CC2CCCCC2[N:1]=C1 |atomProp:9.molAtomMapNumber.1|"


def test_to_smiles_fail():
    smiles = dm.to_smiles(55, allow_to_fail=False)
    assert smiles == None

    # NOTE(hadim): ideally you want to catch only `Boost.Python.ArgumentError` here.
    with pytest.raises(Exception):
        dm.to_smiles(55, allow_to_fail=True)


def test_from_df_pop_mol_column():
    df = dm.data.freesolv().iloc[:10]  # type: ignore
    mols = [dm.to_mol(smiles) for smiles in df["smiles"]]

    df: pd.DataFrame = dm.to_df(mols, mol_column="mol")  # type: ignore
    df["dummy"] = "hello"

    # test with provided mol column
    mols = dm.from_df(df.copy(), mol_column="mol")
    assert set(mols[0].GetPropsAsDict().keys()) == {"smiles", "dummy"}

    # test with automatic mol column detection
    mols = dm.from_df(df.copy())
    assert set(mols[0].GetPropsAsDict().keys()) == {"smiles", "dummy"}


def test_non_standard_inchi():
    mol1 = dm.to_mol("N=C(N)O")
    mol2 = dm.to_mol("NC(N)=O")
    mol3 = dm.to_mol("c1ccccc1")
    mol4 = dm.to_mol("c1cccc([*])c1")
    # with a standard inchi, both molecules should be identical
    assert dm.to_inchi(mol1) == "InChI=1S/CH4N2O/c2-1(3)4/h(H4,2,3,4)"
    assert dm.to_inchi(mol2) == "InChI=1S/CH4N2O/c2-1(3)4/h(H4,2,3,4)"
    assert dm.to_inchikey(mol1) == "XSQUKJJJFZCRTK-UHFFFAOYSA-N"
    assert dm.to_inchikey(mol2) == "XSQUKJJJFZCRTK-UHFFFAOYSA-N"

    # with additional layers, inchi values should be different
    assert dm.to_inchi_non_standard(mol1) == "InChI=1/CH4N2O/c2-1(3)4/h(H4,2,3,4)/f/h2,4H,3H2/b2-1?"
    assert dm.to_inchi_non_standard(mol2) == "InChI=1/CH4N2O/c2-1(3)4/h(H4,2,3,4)/f/h2-3H2"
    assert dm.to_inchikey_non_standard(mol1) == "XSQUKJJJFZCRTK-ZIALIONUNA-N"
    assert dm.to_inchikey_non_standard(mol2) == "XSQUKJJJFZCRTK-UBUOBULFNA-N"

    # for molecule without movile hydrogens the inchi and inchikey will still be different
    # because of the absence of the letter `S` in the prefix
    assert dm.to_inchi_non_standard(mol3) == "InChI=1/C6H6/c1-2-4-6-5-3-1/h1-6H"
    assert dm.to_inchi(mol3) == "InChI=1S/C6H6/c1-2-4-6-5-3-1/h1-6H"
    assert dm.to_inchikey_non_standard(mol3) == "UHOVQNZJYSORNB-UHFFFAOYNA-N"
    assert dm.to_inchikey(mol3) == "UHOVQNZJYSORNB-UHFFFAOYSA-N"

    # let's check we can reload the correct molecule from the non standard inchi
    # The SMILES is unlikely to change with a simple molecule like that
    mol1_reloaded = dm.from_inchi(dm.to_inchi_non_standard(mol1))
    mol2_reloaded = dm.from_inchi(dm.to_inchi_non_standard(mol2))

    assert dm.to_smiles(mol1_reloaded) == "N=C(N)O"
    assert dm.to_smiles(mol2_reloaded) == "NC(N)=O"

    # inchikey computation should return None on query
    assert dm.to_inchikey(mol4) is None


def test_non_standard_inchi_with_options():
    mol1 = dm.to_mol("C[C@@H]1CCCC[C@@H]1O")
    mol2 = dm.to_mol("C[C@H]1CCCC[C@@H]1O")

    assert dm.to_inchikey(mol1) == "NDVWOBYBJYUSMF-RQJHMYQMSA-N"
    assert dm.to_inchikey_non_standard(mol1) == "NDVWOBYBJYUSMF-RQJHMYQMNA-N"
    assert dm.to_inchikey_non_standard(mol1, options=["/SRel"]) == "NDVWOBYBJYUSMF-JHPDDGAFNA-N"

    assert dm.to_inchikey(mol2) == "NDVWOBYBJYUSMF-BQBZGAKWSA-N"
    assert dm.to_inchikey_non_standard(mol2) == "NDVWOBYBJYUSMF-BQBZGAKWNA-N"
    assert dm.to_inchikey_non_standard(mol2, options=["/SRel"]) == "NDVWOBYBJYUSMF-WZTWBHKBNA-N"


def test_non_standard_inchi_edge_cases():
    assert dm.to_inchi_non_standard(None) is None
    assert dm.to_inchikey_non_standard(None) is None

    # test with wrong query molecules
    assert dm.to_inchi_non_standard("NC([*])C(N)=O") is None
    assert dm.to_inchikey_non_standard("NC([*])C(N)=O") is None

    # test with empty molecule
    assert dm.to_inchi_non_standard("CCC(C)(C)(C)(C)(C)") is None
    assert dm.to_inchikey_non_standard("CCC(C)(C)(C)(C)(C)") is None

    # test with invalid molecule
    assert dm.to_inchi_non_standard("") is None
    assert dm.to_inchikey_non_standard("") is None

    assert dm.to_inchi_non_standard("NC(N)=O") == "InChI=1/CH4N2O/c2-1(3)4/h(H4,2,3,4)/f/h2-3H2"
    assert dm.to_inchikey_non_standard("N=C(N)O") == "XSQUKJJJFZCRTK-ZIALIONUNA-N"


def test_to_smiles_with_indices():
    mol = dm.to_mol("Cn1c(=O)c2c(ncn2C)n(C)c1=O")

    # NOTE(hadim): not a great test I guess xD
    smiles = dm.to_smiles(mol, with_atom_indices=True)
    smiles = cast(str, smiles)
    assert all([str(i) in smiles for i in range(13)])



from . import utils

from ._viz import to_image

from ._substructure import match_substructure

from ._conformers import conformers

from ._circle_grid import circle_grid
from ._circle_grid import MolsCircleGrid



"""
sanifix4.py
Original code from rdkit [James Davidson]
"""

from rdkit import Chem, RDLogger


logger = RDLogger.logger()


def _FragIndicesToMol(oMol, indices):
    em = Chem.EditableMol(Chem.Mol())

    newIndices = {}
    for i, idx in enumerate(indices):
        em.AddAtom(oMol.GetAtomWithIdx(idx))
        newIndices[idx] = i

    for i, idx in enumerate(indices):
        at = oMol.GetAtomWithIdx(idx)
        for bond in at.GetBonds():
            if bond.GetBeginAtomIdx() == idx:
                oidx = bond.GetEndAtomIdx()
            else:
                oidx = bond.GetBeginAtomIdx()
            # make sure every bond only gets added once:
            if oidx < idx:
                continue
            em.AddBond(newIndices[idx], newIndices[oidx], bond.GetBondType())
    res = em.GetMol()
    res.ClearComputedProps()
    Chem.GetSymmSSSR(res)
    res.UpdatePropertyCache(False)
    res._idxMap = newIndices
    return res


def _recursivelyModifyNs(mol, matches, indices=None):
    if indices is None:
        indices = []
    res = None
    while len(matches) and res is None:
        tIndices = indices[:]
        nextIdx = matches.pop(0)
        tIndices.append(nextIdx)
        nm = Chem.Mol(mol.ToBinary())
        nm.GetAtomWithIdx(nextIdx).SetNoImplicit(True)
        nm.GetAtomWithIdx(nextIdx).SetNumExplicitHs(1)
        cp = Chem.Mol(nm.ToBinary())
        try:
            Chem.SanitizeMol(cp)
        except ValueError:
            res, indices = _recursivelyModifyNs(nm, matches, indices=tIndices)
        else:
            indices = tIndices
            res = cp
    return res, indices


def AdjustAromaticNs(m, nitrogenPattern="[n&D2&H0;r5,r6]"):
    """
    default nitrogen pattern matches Ns in 5 rings and 6 rings in order to be able
    to fix: O=c1ccncc1
    """
    Chem.GetSymmSSSR(m)
    m.UpdatePropertyCache(False)

    # break non-ring bonds linking rings:
    em = Chem.EditableMol(m)
    linkers = m.GetSubstructMatches(Chem.MolFromSmarts("[r]!@[r]"))
    plsFix = set()
    for a, b in linkers:
        em.RemoveBond(a, b)
        plsFix.add(a)
        plsFix.add(b)
    nm = em.GetMol()
    for at in plsFix:
        at = nm.GetAtomWithIdx(at)
        if at.GetIsAromatic() and at.GetAtomicNum() == 7:
            at.SetNumExplicitHs(1)
            at.SetNoImplicit(True)

    # build molecules from the fragments:
    fragLists = Chem.GetMolFrags(nm)
    frags = [_FragIndicesToMol(nm, x) for x in fragLists]

    # loop through the fragments in turn and try to aromatize them:
    ok = True
    for i, frag in enumerate(frags):
        cp = Chem.Mol(frag)
        try:
            Chem.SanitizeMol(cp)
        except ValueError:
            matches = [x[0] for x in frag.GetSubstructMatches(Chem.MolFromSmarts(nitrogenPattern))]
            lres, indices = _recursivelyModifyNs(frag, matches)
            if not lres:
                # print 'frag %d failed (%s)'%(i,str(fragLists[i]))
                ok = False
                break
            else:
                revMap = {}
                for k, v in frag._idxMap.items():
                    revMap[v] = k
                for idx in indices:
                    oatom = m.GetAtomWithIdx(revMap[idx])
                    oatom.SetNoImplicit(True)
                    oatom.SetNumExplicitHs(1)
    if not ok:
        return None
    return m


def sanifix(m):
    if m is None:
        return None
    try:
        m.UpdatePropertyCache(False)
        cp = Chem.Mol(m.ToBinary())
        Chem.SanitizeMol(cp)
        return cp
    except ValueError as e:
        logger.debug(f"{Chem.MolToSmiles(m)} failed due to {e}")
        try:
            m = AdjustAromaticNs(m)
            if m is not None:
                Chem.SanitizeMol(m)
            return m
        except Exception as ee:
            logger.debug(f"{Chem.MolToSmiles(m)} failed due to {ee}")
            return None
    except RuntimeError as e:
        logger.debug(f"{Chem.MolToSmiles(m)} failed due to {e}")
        logger.info(f"The faulty smiles is: {Chem.MolToSmiles(m)}")
        raise e



import pytest

import numpy as np
import datamol as dm
import datamol.utils.testing


def test_pdist():
    smiles_list = ["CC(=O)Oc1ccccc1C(=O)O", "C1OC1CC", "c1cc2ccccc2cc1"]
    mols = [dm.to_mol(smiles) for smiles in smiles_list]

    dist_mat = dm.pdist(mols)

    assert dist_mat.shape == (3, 3)
    assert dist_mat.sum() == 5.6757105943152455

    dist_mat = dm.pdist(mols, n_jobs=None)

    assert dist_mat.shape == (3, 3)
    assert dist_mat.sum() == 5.6757105943152455


def test_pdist_condensed():
    smiles_list = ["CC(=O)Oc1ccccc1C(=O)O", "C1OC1CC", "c1cc2ccccc2cc1"]
    mols = [dm.to_mol(smiles) for smiles in smiles_list]

    dist_mat = dm.pdist(mols, squareform=False)

    assert dist_mat.shape == (3,)
    assert dist_mat.sum() == 2.8378552971576227


def test_cdist():
    smiles_list1 = ["CC(=O)Oc1ccccc1C(=O)O", "C1OC1CC", "c1cc2ccccc2cc1"]
    mols1 = [dm.to_mol(smiles) for smiles in smiles_list1]

    smiles_list2 = [
        "COc1cc(Nc2ncc(Cl)c(-c3cccc(CC#N)c3)n2)ccc1N1CCN(C)CC1",
        "ON=C(O)CCCCCN=C(O)C=C1c2ccccc2-c2ccccc21",
        "COc1ccc(CCc2nnc(-c3ccc4nc[nH]c4c3)o2)cc1Cl",
    ]
    mols2 = [dm.to_mol(smiles) for smiles in smiles_list2]

    dist_mat = dm.cdist(mols1, mols2)

    assert dist_mat.shape == (3, 3)
    assert np.isclose(dist_mat.mean(), 0.9416270180919872)


def test_cdist_chunked():
    smiles_list1 = ["CC(=O)Oc1ccccc1C(=O)O", "C1OC1CC", "c1cc2ccccc2cc1"]
    mols1 = [dm.to_mol(smiles) for smiles in smiles_list1]

    smiles_list2 = [
        "COc1cc(Nc2ncc(Cl)c(-c3cccc(CC#N)c3)n2)ccc1N1CCN(C)CC1",
        "ON=C(O)CCCCCN=C(O)C=C1c2ccccc2-c2ccccc21",
        "COc1ccc(CCc2nnc(-c3ccc4nc[nH]c4c3)o2)cc1Cl",
    ]
    mols2 = [dm.to_mol(smiles) for smiles in smiles_list2]

    d1 = dm.cdist(mols1, mols2, distances_chunk=True)
    d2 = dm.cdist(mols1, mols2, distances_chunk=False)

    assert d1.shape == d2.shape
    assert np.allclose(d1, d2)


def test_cdist_pdist_consistent():
    smiles_list1 = ["CC(=O)Oc1ccccc1C(=O)O", "C1OC1CC", "c1cc2ccccc2cc1"]
    mols1 = [dm.to_mol(smiles) for smiles in smiles_list1]

    dist_mat = dm.cdist(mols1, mols1)
    dist_mat2 = dm.pdist(mols1)

    assert np.isclose(dist_mat.mean(), dist_mat2.mean())
    assert np.allclose(dist_mat, dist_mat2)


def test_cdist_pdist_invalid_input():
    smiles_list = ["CC(=O)Oc1ccccc1C(=O)O", "C1OC1CC", "c1cc2ccccc2cc1", "dsdsdsd"]

    with pytest.raises(ValueError):
        dm.similarity.cdist(smiles_list, smiles_list)

    with pytest.raises(ValueError):
        dm.similarity.pdist(smiles_list)


def test_datamol_pdist_same_as_rdkit():
    smiles_list = [
        "COc1cc(Nc2ncc(Cl)c(-c3cccc(CC#N)c3)n2)ccc1N1CCN(C)CC1",
        "ON=C(O)CCCCCN=C(O)C=C1c2ccccc2-c2ccccc21",
        "COc1ccc(CCc2nnc(-c3ccc4nc[nH]c4c3)o2)cc1Cl",
    ]

    dist_mat = dm.similarity.pdist(smiles_list)
    dist_mat_rdkit = datamol.utils.testing.pdist_rdkit(smiles_list)

    assert np.allclose(dist_mat, dist_mat_rdkit)


def test_datamol_cdist_same_as_rdkit():
    smiles_list = [
        "COc1cc(Nc2ncc(Cl)c(-c3cccc(CC#N)c3)n2)ccc1N1CCN(C)CC1",
        "ON=C(O)CCCCCN=C(O)C=C1c2ccccc2-c2ccccc21",
        "COc1ccc(CCc2nnc(-c3ccc4nc[nH]c4c3)o2)cc1Cl",
    ]

    smiles_list2 = ["CC(=O)Oc1ccccc1C(=O)O", "C1OC1CC", "c1cc2ccccc2cc1"]

    dist_mat = dm.similarity.cdist(smiles_list, smiles_list2)
    dist_mat_rdkit = datamol.utils.testing.cdist_rdkit(smiles_list, smiles_list2)

    assert np.allclose(dist_mat, dist_mat_rdkit)



import datamol as dm
import numpy as np

REACTION_BLOCK = """$RXN

      RDKit

  2  1
$MOL

     RDKit          2D

  2  1  0  0  0  0  0  0  0  0999 V2000
    0.0000    0.0000    0.0000 *   0  0  0  0  0  0  0  0  0  0  0  0
    1.2990    0.7500    0.0000 *   0  0  0  0  0  0  0  0  0  1  0  0
  1  2  6  0
V    1 [1*]
V    2 [*:1]
M  END
$MOL

     RDKit          2D

  2  1  0  0  0  0  0  0  0  0999 V2000
    0.0000    0.0000    0.0000 *   0  0  0  0  0  0  0  0  0  0  0  0
    1.2990    0.7500    0.0000 *   0  0  0  0  0  0  0  0  0  2  0  0
  1  2  6  0
V    1 [1*]
V    2 [*:2]
M  END
$MOL

     RDKit          2D

  2  1  0  0  0  0  0  0  0  0999 V2000
    0.0000    0.0000    0.0000 *   0  0  0  0  0  0  0  0  0  1  0  0
    1.2990    0.7500    0.0000 *   0  0  0  0  0  0  0  0  0  2  0  0
  1  2  6  0
V    1 [*:1]
V    2 [*:2]
M  END"""


def test_convert_attach_to_isotope():
    smiles = "O=C(NC[*:1])Cc1ccnc(OCc2ccccc2)c1"

    isotope = dm.reactions.convert_attach_to_isotope(mol_or_smiles=smiles)

    assert dm.to_smiles(isotope, canonical=True) == dm.to_smiles(
        dm.to_mol("O=C(Cc1ccnc(OCc2ccccc2)c1)NC[1*]"), canonical=True
    )

    # same isotope
    dm.reactions.convert_attach_to_isotope(mol_or_smiles=smiles, same_isotope=True)

    isotope2 = dm.reactions.convert_attach_to_isotope(mol_or_smiles=smiles, as_smiles=True)
    assert isotope2 == "O=C(Cc1ccnc(OCc2ccccc2)c1)NC[1*]"


def test_num_attachment_points():
    smiles = "O=C(NC[*:1])Cc1c([*:2])nc(OCc2ccccc2)c1"
    assert dm.reactions.num_attachment_points(smiles) == 2

    assert dm.reactions.num_attachment_points(dm.from_smarts(smiles)) == 2


def test_open_attach_points():
    mol = dm.to_mol("CC")
    smiles_open = dm.to_smiles(
        dm.reactions._attachments.open_attach_points(mol, bond_type=dm.SINGLE_BOND), canonical=True
    )
    assert smiles_open == dm.to_smiles(dm.to_mol("*CC*"), canonical=True)


def test_rxn_from_smarts():
    smarts = "[1*][*:1].[1*][*:2]>>[*:1][*:2]"
    rxn = dm.reactions.rxn_from_smarts(smarts)
    assert len(list(rxn.GetReactants())) == 2


def test_rxn_to_smarts():
    smarts = "[1*][*:1].[1*][*:2]>>[*:1][*:2]"
    rxn = dm.reactions.rxn_from_smarts(smarts)

    smarts2 = dm.reactions.rxn_to_smarts(rxn)
    assert smarts == smarts2


def test_rxn_from_block():
    rxn = dm.reactions.rxn_from_block(REACTION_BLOCK)
    assert rxn.GetNumProductTemplates() == 1
    assert rxn.GetNumReactantTemplates() == 2


def test_rxn_from_block_file(tmp_path):
    block_path = tmp_path / "block.block"

    with open(block_path, "w") as f:
        f.write(REACTION_BLOCK)

    rxn = dm.reactions.rxn_from_block_file(block_path)
    assert rxn.GetNumProductTemplates() == 1
    assert rxn.GetNumReactantTemplates() == 2


def test_rxn_to_block():
    rxn = dm.reactions.rxn_from_smarts("[*:1][*:2]>>[*:1].[*:2]")
    block = dm.reactions.rxn_to_block(rxn)
    rxn2 = dm.reactions.rxn_from_block(block)
    assert dm.reactions.rxn_to_smarts(rxn2) == dm.reactions.rxn_to_smarts(rxn)


def test_rxn_to_block_file(tmp_path):
    block_path = tmp_path / "block.block"
    rxn = dm.reactions.rxn_from_smarts("[*:1][*:2]>>[*:1].[*:2]")
    dm.reactions.rxn_to_block_file(rxn, block_path)

    rxn2 = dm.reactions.rxn_from_block_file(block_path)
    assert dm.reactions.rxn_to_smarts(rxn2) == dm.reactions.rxn_to_smarts(rxn)


def test_is_reaction_ok():
    smarts = "[1*][*:1].[1*][*:2]>>[*:1][*:2]"
    rxn = dm.reactions.rxn_from_smarts(smarts)
    assert dm.reactions.is_reaction_ok(rxn) == True


def test_is_reaction_ok_with_logs(caplog):
    smarts = "[1*][*:1].[1*][*:2]>>[*:1][*:2]"
    rxn = dm.reactions.rxn_from_smarts(smarts)
    assert dm.reactions.is_reaction_ok(rxn, enable_logs=True) == True

    assert "Number of" in caplog.text


def test_apply_reaction():
    # 1 attach point
    rxn = dm.reactions.rxn_from_smarts("[1*][*:1].[1*][*:2]>>[*:1][*:2]")
    frag = dm.to_mol("CC(O[1*])C(F)(F)F")
    scf = dm.to_mol("O=C(NC[1*])NCc1ccnc(OCc2ccccc2)c1")
    prod = dm.reactions.apply_reaction(
        rxn=rxn, reactants=(scf, frag), product_index=0, single_product_group=True
    )
    assert len([prod]) == 1
    assert isinstance(prod, dm.Mol) == True

    prod = dm.reactions.apply_reaction(
        rxn=rxn, reactants=(scf, frag), single_product_group=False, as_smiles=True
    )
    assert len(prod) >= 1
    assert isinstance(prod, list) == True


def test_apply_reaction_invalid():
    # 1 attach point
    rxn = dm.reactions.rxn_from_smarts("[C:1](=[O:2])O.[N:3]>>[C:1](=[O:2])[N:3]")
    frag = dm.to_mol("CC(O)C(F)(F)F")
    scf = dm.to_mol("O=C(NC)NCc1ccnc(OCc2cc[2*]ccc2)c1")
    prod = dm.reactions.apply_reaction(
        rxn=rxn, reactants=(scf, frag), product_index=0, single_product_group=True
    )
    assert isinstance(prod, list)
    assert len(prod) == 0


def test_can_react():
    rxn = dm.reactions.rxn_from_smarts(
        "[*:1]O[C:2](=[O])[#6:3].[Nh:4]>>[N:4][C:2](=[O])[#6:3].[*:1][Oh]"
    )
    mol = dm.to_mol("Nc1ccccc1")
    assert dm.reactions.can_react(rxn, mol) == True

    mol = dm.to_mol("CC")
    assert dm.reactions.can_react(rxn, mol) == False


def test_inverse_reaction():
    rxn = dm.reactions.rxn_from_smarts("[1*][*:1].[1*][*:2]>>[*:1][*:2]")
    rxn_r = dm.reactions.inverse_reaction(rxn)
    assert len(list(rxn_r.GetReactants())) == 1
    assert len(list(rxn_r.GetProducts())) == 2


def test_select_reaction_output():
    smiles = (
        (
            "Cc1cnc(CNc(cccc2-c3cn(CC(C4)CC4O)c4ncnc(N)c34)c2F)s1",
            "Cc1cnc(CNc(cccc2-c3cn(C(C4)CC4O)c4ncnc(N)c34)c2F)s1",
        ),
        (
            "Nc1c(c(-c2cccc(NCc3nccs3)c2F)cn2Cc3cn(C4CC5(COCC5)OCC4)nn3)c2ncn1",
            "C[C@@H](c1cn(CC2CCCCC2)nn1)N(Cc1nccs1)c(cccc1-c2c[nH]c3ncnc(N)c23)c1F",
        ),
        (
            "Nc1c(c(-c2cccc(NCc3nccs3)c2F)cn2Cc3cn(CCN4CCOCC4)nn3)c2ncn1",
            "Cc1cnc(C(Nc(cccc2-c3cn(C[C@H](C4)C[C@@H]4O)c4ncnc(N)c34)c2F)=O)s1",
        ),
    )
    product = tuple(tuple([dm.to_mol(p[0]), dm.to_mol(p[1])]) for p in smiles)
    prod = dm.reactions.select_reaction_output(
        product,
        product_index=None,
        single_product_group=False,
        rm_attach=True,
        as_smiles=False,
        sanitize=True,
    )
    assert len(prod) == len(smiles)

    product = tuple(tuple([dm.to_mol(p[0]), dm.to_mol(p[1])]) for p in smiles)
    prod = dm.reactions.select_reaction_output(
        product,
        product_index=[0, 1],
        single_product_group=False,
        rm_attach=True,
        as_smiles=False,
        sanitize=True,
    )
    assert len(prod) == len(smiles)

    product = tuple(tuple([dm.to_mol(p[0]), dm.to_mol(p[1])]) for p in smiles)
    prod = dm.reactions.select_reaction_output(
        product,
        product_index=1,
        single_product_group=True,
        rm_attach=True,
        as_smiles=False,
        sanitize=True,
    )
    assert isinstance(prod, dm.mol.Mol) == True


def test_select_reaction_output_invalid():
    smiles = (
        (
            "Cc1cnc(CNc(cccc2-c3cn(CC(C4)CC4O)c4ncnc(N)c34)c2F)s1",
            "ABCDEFG",  # invalid mol
        ),
        (
            "HIJKLMN",  # invalid mol
            "C[C@@H](c1cn(CC2CCCCC2)nn1)N(Cc1nccs1)c(cccc1-c2c[nH]c3ncnc(N)c23)c1F",
        ),
        (
            "Nc1c(c(-c2cccc(NCc3nccs3)c2F)cn2Cc3cn(CCN4CCOCC4)nn3)c2ncn1",
            "Cc1cnc(C(Nc(cccc2-c3cn(C[C@H](C4)C[C@@H]4O)c4ncnc(N)c34)c2F)=O)s1",
        ),
    )

    product = tuple(tuple([dm.to_mol(p[0]), dm.to_mol(p[1])]) for p in smiles)
    prod = dm.reactions.select_reaction_output(
        product=product,
        product_index=1,
        single_product_group=False,
        rm_attach=True,
        as_smiles=True,
        sanitize=True,
    )
    assert prod[0] is None
    assert isinstance(prod[1], str)
    assert isinstance(prod[2], str)



from typing import Union

import pandas as pd


from .. import Mol

from ..descriptors.descriptors import clogp
from ..descriptors.descriptors import mw
from ..descriptors.descriptors import n_rotatable_bonds
from ..descriptors.descriptors import n_aromatic_atoms_proportion


_ESOL_INTERCEPT = 0.26121066137801696
_ESOL_COEF = {
    "mw": -0.0066138847738667125,
    "clogp": -0.7416739523408995,
    "n_rotatable_bonds": 0.003451545565957996,
    "n_aromatic_atoms_proportion": -0.42624840441316975,
}


def esol(mol: Mol):
    """Compute the solubility descriptor ESOL.

    Note that the intermediate descriptors will be computed on-the-fly. If you prefer
    precomputing those then you can use `esol_from_data`.

    Source: https://github.com/PatWalters/solubility/blob/d1536c58afe5e0e7ac4c96e2ffef496d5b98664b/esol.py
    """

    esol = (
        _ESOL_INTERCEPT
        + _ESOL_COEF["clogp"] * clogp(mol)
        + _ESOL_COEF["mw"] * mw(mol)
        + _ESOL_COEF["n_rotatable_bonds"] * n_rotatable_bonds(mol)
        + _ESOL_COEF["n_aromatic_atoms_proportion"] * n_aromatic_atoms_proportion(mol)
    )

    return esol


def esol_from_data(data: Union[pd.Series, pd.DataFrame, dict]):
    """Compute the solubility descriptor ESOL.

    `data` must contains the following intermediate descriptors:

    - `clogp`: `dm.descriptors.clogp`
    - `mw`: `dm.descriptors.mw`
    - `n_rotatable_bonds`: `dm.descriptors.n_rotatable_bonds`
    - `n_aromatic_atoms_proportion`: `dm.descriptors.n_aromatic_atoms_proportion`

    Source: https://github.com/PatWalters/solubility/blob/d1536c58afe5e0e7ac4c96e2ffef496d5b98664b/esol.py

    Args:
        data: A dataframe or series containing the intermediate descriptors.
    """

    esol = (
        _ESOL_INTERCEPT
        + _ESOL_COEF["clogp"] * data["clogp"]
        + _ESOL_COEF["mw"] * data["mw"]
        + _ESOL_COEF["n_rotatable_bonds"] * data["n_rotatable_bonds"]
        + _ESOL_COEF["n_aromatic_atoms_proportion"] * data["n_aromatic_atoms_proportion"]
    )

    return esol
